from __future__ import annotations

# standard libraries
import gettext
import typing

# third party libraries
# None

# local libraries
from nion.swift import Panel
from nion.swift.model import ApplicationData
from nion.ui import Declarative

if typing.TYPE_CHECKING:
    from nion.swift import DocumentController
    from nion.swift.model import Persistence
    from nion.utils import StructuredModel

_ = gettext.gettext


class SessionHandler(Declarative.Handler):
    def __init__(self, model: StructuredModel.RecordModel) -> None:
        super().__init__()
        self.session_model = model
        self.ui_view = self.__make_ui()

    def __make_ui(self) -> Declarative.UIDescription:
        u = Declarative.DeclarativeUI()

        # ordered from most likely to change to least
        field_descriptions = [
            [_("Sample Area"), _("Sample Area Description"), "sample_area"],
            [_("Sample"), _("Sample Description"), "sample"],
            [_("Task"), _("Task Description"), "task"],
            [_("Microscopist"), _("Microscopist Name(s)"), "microscopist"],
            [_("Instrument"), _("Instrument Description"), "instrument"],
            [_("Site"), _("Site Description"), "site"],
        ]

        field_rows: typing.List[Declarative.UIDescription] = list()

        for field_description in field_descriptions:
            label, tool_tip, key = field_description
            field_line_edit = u.create_line_edit(text=f"@binding(session_model.{key}_model.value)", tool_tip=tool_tip, placeholder_text=tool_tip)
            field_rows.append(u.create_row(u.create_label(text=label, width=120, tool_tip=tool_tip), field_line_edit, spacing=4, margin_horizontal=8, margin_vertical=2))

        return u.create_column(u.create_spacing(4), *field_rows, u.create_row(u.create_stretch(), u.create_label(text=_("Session metadata added to acquired data."), font="italic"), u.create_stretch()), u.create_stretch(), spacing=4)


class SessionPanel(Panel.Panel):

    def __init__(self, document_controller: DocumentController.DocumentController, panel_id: str, properties: Persistence.PersistentDictType) -> None:
        super().__init__(document_controller, panel_id, _("Session"))

        ui_handler = SessionHandler(ApplicationData.get_session_metadata_model())
        self.widget = Declarative.DeclarativeWidget(document_controller.ui, document_controller.event_loop, ui_handler)



"""
A collection of persistence classes.
"""
from __future__ import annotations

# standard libraries
import abc
import copy
import dataclasses
import datetime
import json
import logging
import operator
import re
import traceback
import typing
import uuid
import weakref

# third party libraries
import numpy
import numpy.typing

# local libraries
from nion.utils import Converter
from nion.utils import DateTime
from nion.utils import Event
from nion.utils import Observable
from nion.swift.model import Changes
from nion.swift.model import Utility


PersistentDictType = typing.Dict[str, typing.Any]

_NDArray = numpy.typing.NDArray[typing.Any]


class PersistentContainerType(typing.Protocol):
    def insert_item(self, name: str, before_index: int, item: PersistentObject) -> None: ...
    def remove_item(self, name: str, item: PersistentObject) -> None: ...


_PropertyMakeFn = typing.Callable[[], typing.Any]
_PropertyValidateFn = typing.Callable[[typing.Any], typing.Any]
_PropertyChangedFn = typing.Callable[[str, typing.Any], None]
_PropertyReadFn = typing.Callable[["PersistentProperty", PersistentDictType], typing.Any]
_PropertyWriterFn = typing.Callable[["PersistentProperty", PersistentDictType, typing.Any], None]
_PropertyConverterType = Converter.ConverterLike[typing.Any, typing.Any]  # Utility.CleanValue?
_PersistentObjectFactoryFn = typing.Callable[[typing.Callable[[str], str]], typing.Optional["PersistentObject"]]

_IsEqualType = typing.Callable[[typing.Any, typing.Any], bool]

class PersistentProperty:

    """
        Represents a persistent property.

        converter converts from value to json value
    """

    def __init__(self, name: str, value: typing.Any = None, make: typing.Optional[_PropertyMakeFn] = None,
                 read_only: bool = False, hidden: bool = False, recordable: bool = True,
                 validate: typing.Optional[_PropertyValidateFn] = None,
                 converter: typing.Optional[_PropertyConverterType] = None,
                 changed: typing.Optional[_PropertyChangedFn] = None, key: typing.Optional[str] = None,
                 reader: typing.Optional[_PropertyReadFn] = None,
                 writer: typing.Optional[_PropertyWriterFn] = None,
                 value_type: typing.Optional[typing.Type[typing.Any]] = None,
                 is_equal_fn: typing.Optional[typing.Callable[[typing.Any, typing.Any], bool]] = None) -> None:
        super().__init__()
        self.name = name
        self.key = key if key else name
        self.value: typing.Any = value
        self.make = make
        self.read_only = read_only
        self.hidden = hidden
        self.recordable = recordable
        self.validate = validate
        self.converter = converter
        self.reader = reader
        self.writer = writer
        self.convert_get_fn = typing.cast(typing.Callable[[Utility.DirtyValue], Utility.CleanValue], converter.convert if converter else copy.deepcopy)  # optimization
        self.convert_set_fn = typing.cast(typing.Callable[[Utility.CleanValue], Utility.DirtyValue], converter.convert_back if converter else lambda value: value)  # optimization
        self.changed = changed

        def get_comparable_string(value: typing.Any) -> str:
            d: typing.Dict[str, typing.Any] = dict()
            self.__write_to_dict(d, value)
            return json.dumps(Utility.clean_dict(d), sort_keys=True)

        def is_equal(value1: typing.Any, value2: typing.Any) -> bool:
            return get_comparable_string(value1) == get_comparable_string(value2)

        is_equal_map: typing.Mapping[typing.Any, _IsEqualType] = {
            float: operator.eq,
            str: operator.eq,
            int: operator.eq,
            bool: operator.eq,
            datetime.datetime: operator.eq,
        }

        value_type = value_type or type(value)
        self.is_equal = is_equal_fn if is_equal_fn else is_equal_map.get(value_type, is_equal)

    def close(self) -> None:
        self.make = None
        self.validate = None
        self.converter = None
        self.reader = None
        self.writer = None
        self.convert_get_fn = typing.cast(typing.Any, None)
        self.convert_set_fn = typing.cast(typing.Any, None)
        self.changed = None
        self.is_equal = typing.cast(typing.Any, None)

    def _set_value(self, value: typing.Any, validate: bool) -> tuple[bool, typing.Any]:
        # return whether the value changed
        if self.validate and validate:
            value = self.validate(value)
        else:
            value = copy.deepcopy(value)
        did_change = not self.is_equal(self.value, value)
        self.value = value
        return did_change, value

    def _set_value_call_changed(self, value: typing.Any) -> None:
        if callable(self.changed):
            self.changed(self.name, value)

    def set_value(self, value: typing.Any) -> bool:
        # return whether the value changed
        did_change, value = self._set_value(value, True)
        if did_change:
            self._set_value_call_changed(value)
        return did_change

    @property
    def json_value(self) -> Utility.CleanValue:
        return self.convert_get_fn(self.value)

    @json_value.setter
    def json_value(self, json_value: Utility.CleanValue) -> None:
        self.set_value(self.convert_set_fn(json_value))

    def read_from_dict(self, properties: PersistentDictType) -> None:
        if self.reader:
            value = self.reader(self, properties)
            if value is not None:
                self.set_value(value)
        else:
            if self.key in properties:
                if self.make:
                    value = self.make()
                    value.read_dict(properties[self.key])
                    self.set_value(value)
                else:
                    self.json_value = properties[self.key]

    def __write_to_dict(self, properties: PersistentDictType, value: typing.Any) -> None:
        if self.writer:
            self.writer(self, properties, value)
        else:
            if self.make:
                if value is not None:
                    value_dict = value.write_dict()
                    properties[self.key] = value_dict
                else:
                    properties.pop(self.key, None)  # remove key
            else:
                value = self.convert_get_fn(value)  # json_value
                if value is not None:
                    properties[self.key] = value
                else:
                    properties.pop(self.key, None)  # remove key

    def write_to_dict(self, properties: PersistentDictType) -> None:
        return self.__write_to_dict(properties, self.value)


class PersistentPropertySpecial(PersistentProperty):

    def __init__(self, name: str, value: typing.Any = None, make: typing.Optional[_PropertyMakeFn] = None,
                 read_only: bool = False, hidden: bool = False, recordable: bool = True,
                 validate: typing.Optional[_PropertyValidateFn] = None,
                 converter: typing.Optional[_PropertyConverterType] = None,
                 changed: typing.Optional[_PropertyChangedFn] = None, key: typing.Optional[str] = None,
                 reader: typing.Optional[_PropertyReadFn] = None,
                 writer: typing.Optional[_PropertyWriterFn] = None) -> None:
        super().__init__(name, value, make, read_only, hidden, recordable, validate, converter, changed, key, reader, writer)
        self.__value = value

    @property
    def value(self) -> typing.Any:
        return copy.deepcopy(self.__value)

    @value.setter
    def value(self, value: typing.Any) -> None:
        self.__value = value


class PersistentItem:

    def __init__(self, name: str, factory: _PersistentObjectFactoryFn,
                 item_changed: typing.Optional[typing.Callable[[str, typing.Any, typing.Any], None]] = None,
                 hidden: bool = False) -> None:
        super().__init__()
        self.name = name
        self.factory = factory
        self.item_changed = item_changed
        self.hidden = hidden
        self.value: typing.Any = None
        self.persistent_object_context: typing.Optional[PersistentObjectContext] = None

    def close(self) -> None:
        self.item_changed = None


class PersistentRelationship:

    def __init__(self, name: str,
                 factory: _PersistentObjectFactoryFn,
                 insert: typing.Optional[typing.Callable[[str, int, typing.Any], None]] = None,
                 remove: typing.Optional[typing.Callable[[str, int, typing.Any], None]] = None,
                 key: typing.Optional[str] = None,
                 hidden: bool = False) -> None:
        super().__init__()
        self.name = name
        self.factory = factory
        self.insert = insert
        self.remove = remove
        self.key = key
        self.hidden = hidden
        self.values: typing.List[PersistentObject] = list()
        self.index: typing.Dict[uuid.UUID, PersistentObject] = dict()

    def close(self) -> None:
        self.insert = None
        self.remove = None

    @property
    def storage_key(self) -> str:
        return self.key if self.key else self.name


@dataclasses.dataclass
class ReaderError:
    identifier: str
    exception: Exception
    tb: traceback.StackSummary


class PersistentStorageInterface(typing.Protocol):

    @abc.abstractmethod
    def get_identifier(self) -> str: ...

    @abc.abstractmethod
    def close(self) -> None: ...

    @abc.abstractmethod
    def reset(self) -> None: ...

    @abc.abstractmethod
    def set_root_item(self, item: PersistentObject) -> None: ...

    @abc.abstractmethod
    def unload_item(self, item: PersistentObject) -> None: ...

    @abc.abstractmethod
    def load_properties(self) -> None: ...

    @abc.abstractmethod
    def read_project_properties(self) -> typing.Tuple[PersistentDictType, typing.Sequence[ReaderError]]:
        """Read from storage."""
        ...

    @abc.abstractmethod
    def get_storage_properties(self) -> typing.Optional[PersistentDictType]: ...

    @abc.abstractmethod
    def migrate_to_latest(self) -> None: ...

    @abc.abstractmethod
    def get_properties(self, object: typing.Any) -> typing.Optional[PersistentDictType]: ...

    @abc.abstractmethod
    def insert_relationship_item(self, parent: PersistentObject, name: str, before_index: int, item: PersistentObject) -> None: ...

    @abc.abstractmethod
    def remove_relationship_item(self, parent: PersistentObject, name: str, index: int, item: PersistentObject) -> None: ...

    @abc.abstractmethod
    def load_relationship_item(self, parent: PersistentObject, name: str, before_index: int, item: PersistentObject) -> None: ...

    @abc.abstractmethod
    def unload_relationship_item(self, parent: PersistentObject, item: PersistentObject) -> None: ...

    @abc.abstractmethod
    def load_component_item(self, parent: PersistentObject, name: str, item: PersistentObject) -> None: ...

    @abc.abstractmethod
    def set_component_item(self, parent: PersistentObject, name: str, item: typing.Optional[PersistentObject]) -> None: ...

    @abc.abstractmethod
    def set_property(self, object: PersistentObject, name: str, value: typing.Any, delayed: bool = False) -> None: ...

    @abc.abstractmethod
    def clear_property(self, object: PersistentObject, name: str) -> None: ...

    @abc.abstractmethod
    def read_external_data(self, item: PersistentObject, name: str) -> typing.Any: ...

    @abc.abstractmethod
    def write_external_data(self, item: PersistentObject, name: str, value: _NDArray) -> None: ...

    @abc.abstractmethod
    def reserve_external_data(self, item: PersistentObject, name: str, data_shape: typing.Tuple[int, ...], data_dtype: numpy.typing.DTypeLike) -> None: ...

    @abc.abstractmethod
    def enter_write_delay(self, object: PersistentObject) -> None: ...

    @abc.abstractmethod
    def exit_write_delay(self, object: PersistentObject) -> None: ...

    @abc.abstractmethod
    def is_write_delayed(self, item: PersistentObject) -> bool: ...

    @abc.abstractmethod
    def rewrite_item(self, item: PersistentObject) -> None: ...

    @abc.abstractmethod
    def restore_item(self, data_item_uuid: uuid.UUID) -> typing.Optional[PersistentDictType]: ...

    @abc.abstractmethod
    def prune(self) -> None: ...

    @abc.abstractmethod
    def enter_transaction(self) -> None: ...

    @abc.abstractmethod
    def exit_transaction(self) -> None: ...

    @abc.abstractmethod
    def _get_persistence_write_count(self, item: PersistentObject) -> typing.Optional[int]: ...


class PersistentObjectContext:
    """Provides a common context to track available persistent objects.

    All objects participating in this context should register and unregister themselves with this context at appropriate
    times.

    Other objects can listen to the registration_event to know when a specific object is registered or unregistered.
    """

    def __init__(self) -> None:
        # Python 3.9+: weakref typing
        self.__objects: typing.Dict[PersistentObjectSpecifier, typing.Any] = dict()
        self.registration_event = Event.Event()
        self.__registration_changed_map: typing.Dict[uuid.UUID, typing.Dict[typing.Any, typing.Callable[[typing.Optional[PersistentObject], typing.Optional[PersistentObject]], None]]] = dict()

    def unregister_registration_changed_fn(self, uuid_: uuid.UUID, key: typing.Any) -> None:
        registration_changed_key_map = self.__registration_changed_map.get(uuid_, dict())
        registration_changed_key_map.pop(key)
        if not registration_changed_key_map:
            self.__registration_changed_map.pop(uuid_, None)

    def register_registration_changed_fn(self, uuid_: uuid.UUID, key: typing.Any, registration_changed_fn: typing.Callable[[typing.Optional[PersistentObject], typing.Optional[PersistentObject]], None]) -> None:
        registration_changed_key_map = self.__registration_changed_map.setdefault(uuid_, dict())
        registration_changed_key_map[key] = registration_changed_fn

    def register(self, object: PersistentObject) -> None:
        # print(f"register {object} {item_specifier.write()} {len(self.__objects) + 1}")
        # assert item_specifier not in self.__objects
        item_specifier = object.item_specifier
        self.__objects[item_specifier] = weakref.ref(object)
        self.registration_event.fire(object, None)
        registration_changed_key_map = self.__registration_changed_map.get(object.uuid, dict())
        for registration_changed_fn in list(registration_changed_key_map.values()):
            if callable(registration_changed_fn):
                registration_changed_fn(object, None)

    def unregister(self, object: PersistentObject) -> None:
        # print(f"unregister {object} {item_specifier.write()} {len(self.__objects) - 1}")
        # assert item_specifier in self.__objects
        item_specifier = object.item_specifier
        if item_specifier in self.__objects:
            self.__objects.pop(item_specifier)
            self.registration_event.fire(None, object)
            registration_changed_key_map = self.__registration_changed_map.get(object.uuid, dict())
            for registration_changed_fn in list(registration_changed_key_map.values()):
                if callable(registration_changed_fn):
                    registration_changed_fn(None, object)

    def get_registered_object(self, item_specifier: PersistentObjectSpecifier) -> typing.Optional[PersistentObject]:
        object_weakref = self.__objects.get(item_specifier, None)
        return object_weakref() if object_weakref else None

    def create_item_reference(self, item_uuid: typing.Optional[uuid.UUID] = None, item: typing.Optional[PersistentObject] = None) -> PersistentObjectReference:
        item_specifier = PersistentObjectSpecifier(item_uuid) if item_uuid else None
        return PersistentObjectReference(self, item_specifier, item)


_SpecifierType = typing.Union[typing.Mapping[str, typing.Any], str, uuid.UUID]
_SpecifierDictType = typing.Union[PersistentDictType, str, uuid.UUID]


class PersistentObjectSpecifiable(typing.Protocol):
    @property
    def uuid(self) -> uuid.UUID: raise NotImplementedError()


class PersistentObjectSpecifier:

    def __init__(self, item_uuid_x: typing.Optional[uuid.UUID]) -> None:
        self.__item_uuid = item_uuid_x

    def __hash__(self) -> typing.Any:
        return hash(self.__item_uuid)

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, self.__class__):
            return self.__item_uuid == other.__item_uuid
        return False

    @property
    def item_uuid(self) -> typing.Optional[uuid.UUID]:
        return self.__item_uuid

    def write(self) -> _SpecifierDictType:
        return str(self.item_uuid)


def read_persistent_specifier(d: typing.Optional[_SpecifierType]) -> typing.Optional[PersistentObjectSpecifier]:
    if isinstance(d, str):
        return PersistentObjectSpecifier(uuid.UUID(d))
    elif isinstance(d, uuid.UUID):
        return PersistentObjectSpecifier(d)
    elif isinstance(d, dict) and "item_uuid" in d:
        return PersistentObjectSpecifier(uuid.UUID(d["item_uuid"]))
    elif isinstance(d, dict) and "uuid" in d:
        return PersistentObjectSpecifier(uuid.UUID(d["uuid"]))
    return None


def write_persistent_specifier(item_uuid: uuid.UUID) -> _SpecifierDictType:
    return str(item_uuid)


PO = typing.TypeVar('PO')

class PersistentObjectProxy(typing.Generic[PO]):
    count = 0  # useful for detecting leaks in tests

    def __init__(self, persistent_object: PersistentObject, item_specifier: typing.Optional[PersistentObjectSpecifier], item: typing.Optional[PersistentObject]) -> None:
        PersistentObjectProxy.count += 1
        self.__persistent_object = persistent_object
        self.__item_specifier = item_specifier if item_specifier else PersistentObjectSpecifier(item.uuid) if item else None
        self.__item = item
        self.__persistent_object_context: typing.Optional[PersistentObjectContext] = None
        self.__registered_change_uuid: typing.Optional[uuid.UUID] = None
        self.on_item_registered: typing.Optional[typing.Callable[[PO], None]] = None
        self.on_item_unregistered: typing.Optional[typing.Callable[[PO], None]] = None
        self.__persistent_object_context_changed_listener = persistent_object.persistent_object_context_changed_event.listen(self.__persistent_object_context_changed)
        self.__persistent_object_context_changed()

    def close(self) -> None:
        if self.__persistent_object_context_changed_listener:
            self.__persistent_object_context_changed_listener.close()
            self.__persistent_object_context_changed_listener = typing.cast(typing.Any, None)
        self.__item = None
        self.__item_specifier = None
        self.__update_persistent_object_context()
        self.__persistent_object = typing.cast(typing.Any, None)
        self.on_item_registered = None
        self.on_item_unregistered = None
        PersistentObjectProxy.count -= 1

    @property
    def item(self) -> typing.Optional[PO]:
        return typing.cast(typing.Optional[PO], self.__item)

    @item.setter
    def item(self, item: typing.Optional[PO]) -> None:
        po_item = typing.cast(PersistentObject, item)
        self.__item = po_item
        self.__item_specifier = PersistentObjectSpecifier(po_item.uuid) if item else None
        self.__persistent_object_context_changed()

    @property
    def item_specifier(self) -> typing.Optional[PersistentObjectSpecifier]:
        return self.__item_specifier

    @item_specifier.setter
    def item_specifier(self, item_specifier: PersistentObjectSpecifier) -> None:
        self.__item_specifier = item_specifier
        self.__item = None
        self.__persistent_object_context_changed()

    def __update_persistent_object_context(self) -> None:
        if self.__persistent_object_context:
            if self.__registered_change_uuid:
                self.__persistent_object_context.unregister_registration_changed_fn(self.__registered_change_uuid, self)
                self.__registered_change_uuid = None
            self.__persistent_object_context = None
        if self.__item_specifier and self.__persistent_object.persistent_object_context:
            self.__persistent_object_context = self.__persistent_object.persistent_object_context
            self.__registered_change_uuid = self.__item_specifier.item_uuid
            if self.__registered_change_uuid:
                self.__persistent_object_context.register_registration_changed_fn(self.__registered_change_uuid, self, self.__change_registration)

    def __change_registration(self, registered_object: typing.Optional[PersistentObject], unregistered_object: typing.Optional[PersistentObject]) -> None:
        if registered_object and not self.__item and self.__item_specifier and registered_object.uuid == self.__item_specifier.item_uuid:
            if self.__persistent_object.persistent_object_context:
                item = self.__persistent_object.persistent_object_context.get_registered_object(self.__item_specifier)
                if item:
                    self.__item = item
                    if callable(self.on_item_registered):
                        self.on_item_registered(typing.cast(PO, registered_object))
        if unregistered_object and unregistered_object == self.__item:
            self.__item = None
            if callable(self.on_item_unregistered):
                self.on_item_unregistered(typing.cast(PO, unregistered_object))

    def __persistent_object_context_changed(self) -> None:
        if self.__persistent_object.persistent_object_context:
            if self.__item_specifier and not self.__item:
                item = self.__persistent_object.persistent_object_context.get_registered_object(self.__item_specifier)
                if item:
                    self.__change_registration(item, None)
        self.__update_persistent_object_context()


class PersistentObjectReference:
    count = 0  # useful for detecting leaks in tests

    def __init__(self, persistent_object_context: typing.Optional[PersistentObjectContext], item_specifier: typing.Optional[PersistentObjectSpecifier], item: typing.Optional[PersistentObject]) -> None:
        PersistentObjectReference.count += 1
        self.__persistent_object_context: typing.Optional[PersistentObjectContext] = None
        self.__item_specifier = item_specifier if item_specifier else PersistentObjectSpecifier(item.uuid) if item else None
        self.__item = item
        self.__registered_change_uuid: typing.Optional[uuid.UUID] = None
        self.on_item_registered: typing.Optional[typing.Callable[[PersistentObject], None]] = None
        self.on_item_unregistered: typing.Optional[typing.Callable[[PersistentObject], None]] = None
        self.set_persistent_object_context(persistent_object_context)

    def close(self) -> None:
        self.set_persistent_object_context(None)
        self.__item = None
        self.__item_specifier = None
        self.__persistent_object = None
        self.on_item_registered = None
        self.on_item_unregistered = None
        PersistentObjectReference.count -= 1

    @property
    def item(self) -> typing.Optional[PersistentObject]:
        return self.__item

    @item.setter
    def item(self, item: typing.Optional[PersistentObject]) -> None:
        self.__item = item
        self.__item_specifier = PersistentObjectSpecifier(item.uuid) if item else None
        self.__persistent_object_context_changed()

    @property
    def item_specifier(self) -> PersistentObjectSpecifier | None:
        return self.__item_specifier

    @item_specifier.setter
    def item_specifier(self, item_specifier: PersistentObjectSpecifier | None) -> None:
        self.__item_specifier = item_specifier
        self.__item = None
        self.__persistent_object_context_changed()

    def set_persistent_object_context(self, persistent_object_context: typing.Optional[PersistentObjectContext]) -> None:
        if self.__persistent_object_context:
            if self.__registered_change_uuid:  # use 2nd line to satisfy PyCharm type checker
                self.__persistent_object_context.unregister_registration_changed_fn(self.__registered_change_uuid, self)
                self.__registered_change_uuid = None
        self.__persistent_object_context = persistent_object_context
        if self.__persistent_object_context:
            if self.__item_specifier:
                item_uuid = self.__item_specifier.item_uuid
                assert item_uuid is not None
                self.__registered_change_uuid = item_uuid
                self.__persistent_object_context.register_registration_changed_fn(item_uuid, self, self.__change_registration)
        self.__persistent_object_context_changed()

    def __change_registration(self, registered_object: typing.Optional[PersistentObject], unregistered_object: typing.Optional[PersistentObject]) -> None:
        if registered_object and not self.__item and self.__item_specifier and registered_object.uuid == self.__item_specifier.item_uuid:
            if self.__persistent_object_context:
                item = self.__persistent_object_context.get_registered_object(self.__item_specifier)
                if item:
                    self.__item = item
                    if callable(self.on_item_registered):
                        self.on_item_registered(registered_object)
        if unregistered_object and unregistered_object == self.__item:
            self.__item = None
            if callable(self.on_item_unregistered):
                self.on_item_unregistered(unregistered_object)

    def __persistent_object_context_changed(self) -> None:
        if self.__persistent_object_context:
            if self.__item_specifier and not self.__item:
                item = self.__persistent_object_context.get_registered_object(self.__item_specifier)
                if item:
                    self.__change_registration(item, None)


class PersistentObjectParent:
    """ Track the parent of a persistent object. """

    def __init__(self, parent: PersistentObject, relationship_name: typing.Optional[str] = None, item_name: typing.Optional[str] = None) -> None:
        self.__weak_parent = weakref.ref(parent)
        self.relationship_name = relationship_name
        self.item_name = item_name

    @property
    def parent(self) -> typing.Optional[PersistentObject]:
        return self.__weak_parent()



class PersistentObject(Observable.Observable):
    """
        Base class for objects being stored in a PersistentObjectContext.

        Subclasses can define properties, items, and relationships. Changes to those items will
        be persisted to the PersistentObjectContext.

        Keeps track of modified field automatically.

        Properties are single values, items are a one-to-one between objects, relationships are
        one-to-many between objects.

        Properties can have validators, converters, change notifications, and more.
        They are created using the define_property method.

        Items have set notifications and more. They are created using the
        define_item method.

        Relationships have change notifications and more. They are created using
        the define_relationship method.

        Subclasses can set the uuid after init. It should not be changed at other times.

        The persistent_object_context property must be set explicitly for top level objects of the
        PersistentObjectContext. Objects contained in the items and relationships will be have their
        persistent_object_context managed when they are inserted into or removed from another persistent object.

        It is an error to add an object as an item or relationship more than once. Items can be
        removed and added again, though.

        Objects must be able to read from and write themselves to a dict.

        The persistent_object_context property will be valid after (but not during) reading.

        After reading, an object may immediately update itself to a newer version using the persistent object
        context.
    """
    count = 0  # useful for detecting leaks in tests

    def __init__(self) -> None:
        super().__init__()
        PersistentObject.count += 1
        self.__type: typing.Optional[str] = None
        self.__properties: typing.Dict[str, PersistentProperty] = dict()
        self.__items: typing.Dict[str, PersistentItem] = dict()
        self.__relationships: typing.Dict[str, PersistentRelationship] = dict()
        # Python 3.9+: typed weakref
        self.__container_weak_ref: typing.Optional[typing.Any] = None
        self._closed = False
        self.about_to_close_event = Event.Event()
        self._about_to_be_removed = False
        self.about_to_be_removed_event = Event.Event()
        self._is_reading = False
        self.__persistent_object_context: typing.Optional[PersistentObjectContext] = None
        # uuid as a property is too slow, so make it direct
        self.uuid = uuid.uuid4()
        self.__modified_count = 0
        self.modified_state = 0
        self.__modified = DateTime.utcnow()
        self.modified_event = Event.Event()
        self.persistent_object_parent: typing.Optional[PersistentObjectParent] = None
        self.__persistent_storage: typing.Optional[PersistentStorageInterface] = None
        self.persistent_object_context_changed_event = Event.Event()
        self.__item_references: typing.List[PersistentObjectReference] = list()
        # ghost properties can be used to temporarily mark properties so that changes
        # to those properties do not trigger writes to disk, usually for performance reasons.
        self.ghost_properties = typing.Counter[str]()

    def close(self) -> None:
        if self.persistent_storage:
            self.persistent_storage.unload_item(self)
        self.about_to_close_event.fire()
        assert not self._closed
        self._closed = True
        for item_reference in self.__item_references:
            item_reference.close()
        self.close_items()
        self.close_relationships()
        self.__container_weak_ref = None
        self.__persistent_object_context = None
        self.undefine_properties()
        self.undefine_items()
        self.undefine_relationships()
        PersistentObject.count -= 1

    def __deepcopy__(self, memo: typing.Dict[typing.Any, typing.Any]) -> PersistentObject:
        deepcopy = self.__class__()
        deepcopy.deepcopy_from(self, memo)
        memo[id(self)] = deepcopy
        return deepcopy

    def deepcopy_from(self, item: PersistentObject, memo: typing.Dict[typing.Any, typing.Any]) -> None:
        for key in self.__properties.keys():
            self._copy_property(item, key)
        for key in self.__items.keys():
            self.set_item(key, copy.deepcopy(getattr(item, key)))
        for key in self.__relationships.keys():
            for child_item in getattr(item, key):
                self.append_item(key, copy.deepcopy(child_item, memo))

    def _copy_property(self, item: PersistentObject, key: str) -> None:
        value = item._get_persistent_property_value(key)
        new_value = copy.deepcopy(value)
        self._set_persistent_property_value(key, new_value)

    @property
    def container(self) -> typing.Optional[PersistentObject]:
        return self.__container_weak_ref() if self.__container_weak_ref else None

    def insert_model_item(self, container: PersistentContainerType, name: str, before_index: int, item: PersistentObject) -> None:
        """Insert a model item. Let this item's container do it if possible; otherwise do it directly.

        Passing responsibility to this item's container allows the library to easily track dependencies.
        However, if this item isn't yet in the library hierarchy, then do the operation directly.
        """
        if self.container:
            self.container.insert_model_item(container, name, before_index, item)
        else:
            container.insert_item(name, before_index, item)

    def remove_model_item(self, container: PersistentContainerType, name: str, item: PersistentObject, *, safe: bool = False) -> Changes.UndeleteLog:
        """Remove a model item. Let this item's container do it if possible; otherwise do it directly.

        Passing responsibility to this item's container allows the library to easily track dependencies.
        However, if this item isn't yet in the library hierarchy, then do the operation directly.
        """
        if self.container:
            return self.container.remove_model_item(container, name, item, safe=safe)
        else:
            container.remove_item(name, item)
            return Changes.UndeleteLog()

    def about_to_be_inserted(self, container: PersistentObject) -> None:
        assert self.__container_weak_ref is None
        self.__container_weak_ref = weakref.ref(container)

    def about_to_be_removed(self, container: PersistentObject) -> None:
        # called before close and before item is removed from its container
        for item in self.__items.values():
            if item.value:
                item.value.about_to_be_removed(self)
        for relationship in self.__relationships.values():
            for relationship_item in reversed(relationship.values):
                relationship_item.about_to_be_removed(self)
        self.about_to_be_removed_event.fire()
        assert not self._about_to_be_removed
        self._about_to_be_removed = True
        self.__container_weak_ref = None

    @property
    def item_specifier(self) -> PersistentObjectSpecifier:
        return PersistentObjectSpecifier(self.uuid)

    def define_root_context(self) -> None:
        """Define this item to be the root context."""
        self.__persistent_object_context = PersistentObjectContext()

    def set_storage_system(self, storage_system: PersistentStorageInterface) -> None:
        """Set the storage system for this item."""
        storage_system.set_root_item(self)

    def update_storage_system(self) -> None:
        """Update the storage system properties by re-reading from storage.

        Useful when reloading.
        """
        assert self.persistent_storage
        self.persistent_storage.set_root_item(self)

    def update_item_context(self, item: PersistentObject) -> None:
        """Update the context on the item."""
        item.persistent_object_context = self.persistent_object_context

    def persistent_object_context_changed(self) -> None:
        """ Subclasses can override this to be notified when the persistent object context changes. """
        pass

    @property
    def persistent_object_context(self) -> PersistentObjectContext | None:
        """ Return the persistent object context. """
        return self.__persistent_object_context

    @persistent_object_context.setter
    def persistent_object_context(self, persistent_object_context: PersistentObjectContext | None) -> None:
        """ Set the persistent object context and propagate it to contained objects. """
        assert self.__persistent_object_context is None or persistent_object_context is None  # make sure persistent object context is handled cleanly
        old_persistent_object_context = self.__persistent_object_context
        self.__persistent_object_context = persistent_object_context
        for item_reference in self.__item_references:
            item_reference.set_persistent_object_context(persistent_object_context)
        for item in self.__items.values():
            if item.value:
                item.value.persistent_object_context = persistent_object_context
        for relationship in self.__relationships.values():
            for relationship_item in relationship.values:
                if relationship_item:
                    relationship_item.persistent_object_context = persistent_object_context
        if old_persistent_object_context:
            old_persistent_object_context.unregister(self)
        if persistent_object_context:
            self.__item_specifier = self.item_specifier
            persistent_object_context.register(self)
        self.persistent_object_context_changed()
        self.persistent_object_context_changed_event.fire()

    @property
    def persistent_storage(self) -> typing.Optional[PersistentStorageInterface]:
        return self.__persistent_storage

    @persistent_storage.setter
    def persistent_storage(self, persistent_storage: typing.Optional[PersistentStorageInterface]) -> None:
        self.__persistent_storage = persistent_storage

    def define_type(self, type: str) -> None:
        self.__type = type

    def define_property(self, name: str, value: typing.Any = None, make: typing.Optional[_PropertyMakeFn] = None,
                        read_only: bool = False, hidden: bool = False, recordable: bool = True,
                        copy_on_read: bool = False, validate: typing.Optional[_PropertyValidateFn] = None,
                        converter: typing.Optional[_PropertyConverterType] = None,
                        changed: typing.Optional[_PropertyChangedFn] = None, key: typing.Optional[str] = None,
                        reader: typing.Optional[_PropertyReadFn] = None,
                        writer: typing.Optional[_PropertyWriterFn] = None,
                        value_type: typing.Optional[typing.Type[typing.Any]] = None,
                        is_equal_fn: typing.Optional[typing.Callable[[typing.Any, typing.Any], bool]] = None) -> None:
        """ key is what is stored on disk; name is what is used when accessing the property from code. """
        assert hidden
        if read_only:
            print(f"READ ONLY {name}")
        if copy_on_read:
            self.__properties[name] = PersistentPropertySpecial(name, value, make, read_only, hidden, recordable, validate, converter, changed, key, reader, writer)
        else:
            self.__properties[name] = PersistentProperty(name, value, make, read_only, hidden, recordable, validate, converter, changed, key, reader, writer, value_type, is_equal_fn)

    def define_item(self, name: str, factory: _PersistentObjectFactoryFn,
                    changed: typing.Optional[typing.Callable[[str, typing.Any, typing.Any], None]] = None,
                    hidden: bool = False) -> None:
        assert hidden
        self.__items[name] = PersistentItem(name, factory, changed, hidden)

    def define_relationship(self, name: str,
                            factory: _PersistentObjectFactoryFn,
                            insert: typing.Optional[typing.Callable[[str, int, typing.Any], None]] = None,
                            remove: typing.Optional[typing.Callable[[str, int, typing.Any], None]] = None,
                            key: typing.Optional[str] = None, hidden: bool = False) -> None:
        assert hidden
        self.__relationships[name] = PersistentRelationship(name, factory, insert, remove, key, hidden)

    def close_items(self) -> None:
        for item in self.__items.values():
            if item.value:
                if self.persistent_object_context:  # only clear it if it's been set
                    item.value.persistent_object_context = None
                item.value.close()

    def close_relationships(self) -> None:
        for relationship in self.__relationships.values():
            for item in reversed(relationship.values):
                if item:
                    if self.persistent_object_context:  # only clear it if it's been set
                        item.persistent_object_context = None
                    item.close()

    def undefine_properties(self) -> None:
        for property in self.__properties.values():
            property.close()
        self.__properties.clear()

    def undefine_items(self) -> None:
        for item in self.__items.values():
            item.close()
        self.__items.clear()

    def undefine_relationships(self) -> None:
        for relationship in self.__relationships.values():
            relationship.close()
        self.__relationships.clear()

    def get_storage_properties(self) -> typing.Optional[PersistentDictType]:
        """ Return a copy of the properties for the object as a dict. """
        assert self.persistent_storage
        return copy.deepcopy(self.persistent_storage.get_properties(self))

    @property
    def property_names(self) -> typing.Sequence[str]:
        return list(self.__properties.keys())

    @property
    def key_names(self) -> typing.Sequence[str]:
        return [property.key for property in self.__properties.values()]

    @property
    def type(self) -> str:
        return self.__type or str()

    @property
    def modified(self) -> datetime.datetime:
        return self.__modified

    @property
    def modified_count(self) -> int:
        return self.__modified_count

    def _set_modified(self, modified: datetime.datetime) -> None:
        # for testing
        self.__update_modified(modified)
        if self.persistent_object_context:
            self.property_changed("uuid", str(self.uuid))  # dummy write

    def _get_persistence_write_count(self) -> typing.Optional[int]:
        return self.persistent_storage._get_persistence_write_count(self) if self.persistent_storage else None

    @property
    def item_names(self) -> typing.Sequence[str]:
        return list(self.__items.keys())

    @property
    def relationship_names(self) -> typing.Sequence[str]:
        return list(self.__relationships.keys())

    def begin_reading(self) -> None:
        self._is_reading = True

    def read_from_dict(self, properties: PersistentDictType) -> None:
        """ Read from a dict. """
        # uuid is handled specially for performance reasons
        if "uuid" in properties:
            self.uuid = uuid.UUID(properties["uuid"])
        if "modified" in properties:
            self.__modified = datetime.datetime(*list(map(int, re.split(r'[^\d]', properties["modified"]))))  # type: ignore
        # iterate the defined properties
        for key in self.__properties.keys():
            property = self.__properties[key]
            property.read_from_dict(properties)
        for key in self.__items.keys():
            item_dict = typing.cast(PersistentDictType, properties.get(key))
            if item_dict:
                factory = self.__items[key].factory
                # the object has not been constructed yet, but we needs its
                # type or id to construct it. so we need to look it up by key/index/name.
                # to minimize the interface to the factory methods, just pass a closure
                # which looks up by name.
                def lookup_id(name: str, default: typing.Optional[str] = None) -> str:
                    return typing.cast(str, item_dict.get(name, default))

                item = factory(lookup_id)
                if item:
                    # read the item from the dict
                    item.begin_reading()
                    item.read_from_dict(item_dict)
                    self.__set_item(key, item)
                    if self.persistent_storage:
                        self.persistent_storage.load_component_item(self, key, item)
                else:
                    logging.warning(f"Unable to read item {key} {item_dict}")
        for key in self.__relationships.keys():
            storage_key = self.__relationships[key].storage_key
            for item_dict in properties.get(storage_key, list()):
                factory = self.__relationships[key].factory
                # the object has not been constructed yet, but we need its
                # type or id to construct it. so we need to look it up by key/index/name.
                # to minimize the interface to the factory methods, just pass a closure
                # which looks up by name.
                def lookup_id(name: str, default: typing.Optional[str] = None) -> str:
                    return typing.cast(str, item_dict.get(name, default))

                item = factory(lookup_id)
                if item:
                    # read the item from the dict
                    item.begin_reading()
                    item.read_from_dict(item_dict)
                    # insert it into the relationship dict
                    before_index = len(self.__relationships[key].values)
                    self.load_item(key, before_index, item)
                else:
                    logging.warning(f"Unable to read relationship {key} {item_dict}")

    def finish_reading(self) -> None:
        for key in self.__items.keys():
            item = self.__items[key].value
            if item:
                item.finish_reading()
        for key in self.__relationships.keys():
            for item in self.__relationships[key].values:
                item.finish_reading()
        self._is_reading = False

    def write_to_dict(self) -> PersistentDictType:
        """ Write the object to a dict and return it. """
        properties: PersistentDictType = dict()
        if self.__type:
            properties["type"] = self.__type
        properties["uuid"] = str(self.uuid)
        for key in self.__properties.keys():
            property = self.__properties[key]
            property.write_to_dict(properties)
        for key in self.__items.keys():
            item = self.__items[key].value
            if item:
                properties[key] = item.write_to_dict()
        for key in self.__relationships.keys():
            storage_key = self.__relationships[key].storage_key
            items_list = properties.setdefault(storage_key, list())
            for item in self.__relationships[key].values:
                items_list.append(item.write_to_dict())
        return properties

    def _update_persistent_object_context_property(self, name: str) -> None:
        """Update the property given by name in the persistent object context."""
        if self.persistent_object_context:
            properties: PersistentDictType = dict()
            self.__properties[name].write_to_dict(properties)
            if properties:
                for property_key, property_value in properties.items():
                    self.property_changed(property_key, property_value)
            else:
                self.clear_property(name)

    def __update_modified(self, modified: datetime.datetime) -> None:
        self.__modified_count += 1
        self.modified_state += 1
        self.__modified = modified
        self._modified_changed()
        self.modified_event.fire()
        parent = self.persistent_object_parent.parent if self.persistent_object_parent else None
        if parent:
            parent.__update_modified(modified)

    def _modified_changed(self) -> None:
        """Subclasses can override to know when modification date changed."""
        pass

    def _get_persistent_property(self, name: str) -> PersistentProperty:
        """ Subclasses can call this to get a property descriptor. """
        return self.__properties[name]

    def _get_persistent_property_value(self, name: str, default: typing.Any = None) -> typing.Any:
        """ Subclasses can call this to get a hidden property. """
        property = self.__properties.get(name)
        return property.value if property else default

    def _set_persistent_property_value(self, name: str, value: typing.Any, force_update: bool = False, validate: bool = True) -> None:
        """Set a persistent property directly.

         Subclasses can call this to set a hidden property.

         Normally the update is not written to disk unless the value changes. force_update overrides this behavior.
         """
        property = self.__properties[name]
        # in order to send out the change message after the modification date/count is updated,
        # split the set value into _set_value and _set_value_call_changed. this is consistent
        # with the other modification methods.
        did_change, value = property._set_value(value, validate)
        if did_change or force_update:
            self.__update_modified(DateTime.utcnow())
            self._update_persistent_object_context_property(name)
            if did_change:
                property._set_value_call_changed(value)

    def _update_persistent_property(self, name: str, value: typing.Any) -> None:
        """ Subclasses can call this to notify that a custom property was updated. """
        self.__update_modified(DateTime.utcnow())
        if self.persistent_object_context:
            self.property_changed(name, value)

    def _get_relationship_values(self, name: str) -> typing.Sequence[typing.Any]:
        return copy.copy(self.__relationships[name].values)

    def _is_persistent_property_recordable(self, name: str) -> bool:
        property = self.__properties.get(name)
        return (property.recordable and not property.read_only) if (property is not None) else False

    def __set_item(self, name: str, value: typing.Any) -> None:
        """ Set item into item storage and notify. Does not set into persistent storage or update modified. Item can be None. """
        item = self.__items[name]
        old_value = item.value
        item.value = value
        if value:
            value.persistent_object_parent = PersistentObjectParent(self, item_name=name)
            if self.persistent_object_context:
                value.persistent_object_context = self.persistent_object_context
        if item.item_changed:
            item.item_changed(name, old_value, value)

    def get_item(self, name: str) -> typing.Any:
        """ Get item from persistent storage. """
        item = self.__items[name]
        return item.value

    def set_item(self, name: str, value: typing.Any) -> None:
        """ Set item into persistent storage and then into item storage and notify. """
        item = self.__items[name]
        old_value = item.value
        item.value = value
        self.__update_modified(DateTime.utcnow())
        if value:
            value.persistent_object_parent = PersistentObjectParent(self, item_name=name)
        # the persistent_object_parent and item need to be established before
        # calling item_changed.
        if self.persistent_object_context:
            self.item_set(name, value)  # this will also update item's persistent_object_context
            if value:
                value.persistent_object_context = self.persistent_object_context
        if item.item_changed:
            item.item_changed(name, old_value, value)
        if old_value:
            old_value.close()

    def load_item(self, name: str, before_index: int, item: PersistentObject) -> None:
        """ Load item in persistent storage and then into relationship storage, but don't update modified or notify persistent storage. """
        relationship = self.__relationships[name]
        relationship.values.insert(before_index, item)
        relationship.index[item.uuid] = item
        item.about_to_be_inserted(self)
        item.persistent_object_parent = PersistentObjectParent(self, relationship_name=name)
        if self.persistent_storage:
            self.persistent_storage.load_relationship_item(self, name, before_index, item)
        if self.persistent_object_context:  # when item is not top level, self will not have persistent object context
            item.persistent_object_context = self.persistent_object_context
        if relationship.insert:
            relationship.insert(name, before_index, item)

    def unload_item(self, name: str, index: int) -> None:
        """ Unload item from relationship storage and persistent storage, but don't update modified or notify persistent storage. """
        relationship = self.__relationships[name]
        item = relationship.values.pop(index)
        relationship.index.pop(item.uuid)
        item.about_to_be_removed(self)
        if relationship.remove:
            relationship.remove(name, index, item)
        item.persistent_object_context = None
        item.persistent_object_parent = None
        if self.persistent_storage:
            self.persistent_storage.unload_relationship_item(self, item)
        item.close()

    def insert_item(self, name: str, before_index: int, item: PersistentObject) -> None:
        """ Insert item in persistent storage and then into relationship storage and notify. """
        relationship = self.__relationships[name]
        relationship.values.insert(before_index, item)
        relationship.index[item.uuid] = item
        self.__update_modified(DateTime.utcnow())
        item.persistent_object_parent = PersistentObjectParent(self, relationship_name=name)
        # the persistent_object_parent and relationship need to be established before
        # calling item_inserted.
        item.about_to_be_inserted(self)
        if self.persistent_storage:
            self.persistent_storage.insert_relationship_item(self, name, before_index, item)
        if self.persistent_object_context:
            item.persistent_object_context = self.persistent_object_context
        if relationship.insert:
            relationship.insert(name, before_index, item)

    def append_item(self, name: str, item: PersistentObject) -> None:
        """ Append item and append to persistent storage. """
        self.insert_item(name, len(self.__relationships[name].values), item)

    def remove_item(self, name: str, item: PersistentObject) -> None:
        """ Remove item and remove from persistent storage. """
        item.about_to_be_removed(self)
        relationship = self.__relationships[name]
        item_index = relationship.values.index(item)
        relationship.values.remove(item)
        relationship.index.pop(item.uuid)
        self.__update_modified(DateTime.utcnow())
        if relationship.remove:
            relationship.remove(name, item_index, item)
        if self.persistent_object_context:  # only clear if self has a context; it won't if it is still being constructed
            item.persistent_object_context = None
        if self.persistent_storage:
            self.persistent_storage.remove_relationship_item(self, name, item_index, item)  # this will also update item's persistent_object_context
        item.persistent_object_parent = None
        item.close()

    def extend_items(self, name: str, items: typing.Sequence[PersistentObject]) -> None:
        """ Append multiple items and add to persistent storage. """
        for item in items:
            self.append_item(name, item)

    def item_count(self, name: str) -> int:
        """Return the count of items in the relationship specified by name."""
        relationship = self.__relationships[name]
        return len(relationship.values)

    def item_index(self, name: str, item: PersistentObject) -> int:
        """Return the index of item within the relationship specified by name."""
        relationship = self.__relationships[name]
        return relationship.values.index(item)

    def get_item_by_uuid(self, name: str, uuid: uuid.UUID) -> typing.Optional[PersistentObject]:
        """Return the item from the index by uuid."""
        relationship = self.__relationships[name]
        return relationship.index.get(uuid)

    def change_relationship_item_uuid(self, name: str, old_uuid: uuid.UUID, new_uuid: uuid.UUID) -> None:
        """Change the uuid of an item."""
        relationship = self.__relationships[name]
        item = relationship.index.pop(old_uuid)
        relationship.index[new_uuid] = item

    def get_relationship_items(self, name: str) -> typing.List[PersistentObject]:
        """Return the relationship items without making a copy."""
        return self.__relationships[name].values

    def item_set(self, name: str, item: PersistentObject) -> None:
        """ Call this to notify this context that an item with name has been set on the parent. """
        assert self.persistent_storage
        self.persistent_storage.set_component_item(self, name, item)

    def property_changed(self, name: str, value: typing.Any) -> None:
        """ Call this to notify this context that a property with name has changed to value on object. """
        assert self.persistent_storage
        delayed = name in self.ghost_properties.elements()
        self.persistent_storage.set_property(self, name, value, delayed)

    def clear_property(self, name: str) -> None:
        """ Call this to notify this context that a property with name has been removed on object. """
        assert self.persistent_storage
        self.persistent_storage.clear_property(self, name)

    def read_external_data(self, name: str) -> typing.Any:
        """ Call this to notify read external data with name from an item in persistent storage. """
        assert self.persistent_storage
        return self.persistent_storage.read_external_data(self, name)

    def write_external_data(self, name: str, value: typing.Any) -> None:
        """ Call this to notify write external data value with name to an item in persistent storage. """
        assert self.persistent_storage
        self.persistent_storage.write_external_data(self, name, value)

    def reserve_external_data(self, name: str, data_shape: typing.Tuple[int, ...], data_dtype: numpy.typing.DTypeLike) -> None:
        """ Call this to notify reserve external data value with name to an item in persistent storage. """
        assert self.persistent_storage
        self.persistent_storage.reserve_external_data(self, name, data_shape, numpy.dtype(data_dtype))

    def enter_write_delay(self) -> None:
        """ Call this to notify this context that the object should be write delayed. """
        assert self.persistent_storage
        self.persistent_storage.enter_write_delay(self)

    def exit_write_delay(self) -> None:
        """ Call this to notify this context that the object should no longer be write delayed. """
        assert self.persistent_storage
        self.persistent_storage.exit_write_delay(self)

    @property
    def is_write_delayed(self) -> bool:
        assert self.persistent_storage
        return self.persistent_storage.is_write_delayed(self)

    def rewrite(self) -> None:
        """ Call this to write an item that was write delayed. """
        assert self.persistent_storage
        self.persistent_storage.rewrite_item(self)

    def create_item_proxy(self, *, item_uuid: typing.Optional[uuid.UUID] = None, item_specifier: typing.Optional[PersistentObjectSpecifier] = None, item: typing.Optional[PersistentObject] = None) -> PersistentObjectProxy[typing.Any]:
        """Create an item proxy by uuid or directly using the item."""
        item_specifier = item_specifier or (PersistentObjectSpecifier(item_uuid) if item_uuid else None)
        return PersistentObjectProxy(self, item_specifier, item)

    def create_item_reference(self, *, item_uuid: typing.Optional[uuid.UUID] = None, item_specifier: typing.Optional[PersistentObjectSpecifier] = None, item: typing.Optional[PersistentObject] = None) -> PersistentObjectReference:
        """Create an item proxy by uuid or directly using the item."""
        item_specifier = item_specifier or (PersistentObjectSpecifier(item_uuid) if item_uuid else None)
        item_reference = PersistentObjectReference(self.persistent_object_context, item_specifier, item)
        self.__item_references.append(item_reference)
        return item_reference

    def destroy_item_reference(self, item_reference: PersistentObjectReference) -> None:
        self.__item_references.remove(item_reference)

    def resolve_item_specifier(self, item_specifier: PersistentObjectSpecifier) -> typing.Optional[PersistentObject]:
        """Return the resolve item specifier."""
        if self.persistent_object_context:
            return self.persistent_object_context.get_registered_object(item_specifier)
        return None



from __future__ import annotations

# standard libraries
import asyncio
import collections
import dataclasses
import functools
import gettext
import logging
import sys
import threading
import typing
import weakref

# third party libraries
# None

# local libraries
from nion.swift.model import UISettings
from nion.swift.model import Utility
from nion.ui import Application
from nion.ui import CanvasItem
from nion.ui import Declarative
from nion.ui import UserInterface
from nion.utils import Event
from nion.utils import Geometry
from nion.utils import Platform
from nion.utils import ReferenceCounting
from nion.utils import Registry

if typing.TYPE_CHECKING:
    from nion.swift import DocumentController
    from nion.swift.model import Persistence
    from nion.swift.model import UISettings
    from nion.ui import DrawingContext

_DocumentControllerWeakRefType = typing.Callable[[], "DocumentController.DocumentController"]

_ = gettext.gettext


class Panel:
    """
        Represents content within a dock widget. The dock widget owns
        the panel and will invoke close and periodic on it. The dock
        widget expects the widget property to contain the ui content.
    """
    count = 0  # useful for detecting leaks in tests

    @classmethod
    def get_monospace_text_font(cls) -> str:
        if Platform.is_macos():
            return "11px monospace"
        else:
            return "12px monospace"

    @classmethod
    def get_monospace_proportional_line_height(cls) -> float:
        return 1.1

    def __init__(self, document_controller: DocumentController.DocumentController, panel_id: str, display_name: str) -> None:
        Panel.count += 1
        self.__weak_document_controller = typing.cast(_DocumentControllerWeakRefType, weakref.ref(document_controller))
        self.ui = document_controller.ui
        self.panel_id = panel_id
        self.display_name = display_name
        self.widget = typing.cast(UserInterface.Widget, None)  # gets closed by the dock widget
        self.dock_widget = typing.cast(UserInterface.DockWidget, None)

    # subclasses can override to clean up when the panel closes.
    def close(self) -> None:
        if self.dock_widget:
            # added to a dock widget, closing dock widget will close the widget
            self.dock_widget.close()
        else:
            # never added to a dock widget, so explicitly close
            self.widget.close()
        self.widget = typing.cast(UserInterface.Widget, None)  # closed by the dock_widget
        Panel.count -= 1

    def create_dock_widget(self, title: str, positions: typing.Sequence[str], position: str) -> None:
        self.dock_widget = self.document_controller.create_dock_widget(self.widget, self.panel_id, title, positions, position)
        self.dock_widget.on_size_changed = self.size_changed
        self.dock_widget.on_focus_changed = self.focus_changed
        self.dock_widget.on_ui_activity = self.document_controller._register_ui_activity
        self.dock_widget.on_physical_dpi_changed = self.__physical_dpi_changed

    @property
    def document_controller(self) -> DocumentController.DocumentController:
        return self.__weak_document_controller()

    # not thread safe. always call from main thread.
    def periodic(self) -> None:
        self.dock_widget.periodic()

    def show(self) -> None:
        self.dock_widget.show()

    def hide(self) -> None:
        self.dock_widget.hide()

    # tasks can be added in two ways, queued or added
    # queued tasks are guaranteed to be executed in the order queued.
    # added tasks are only executed if not replaced before execution.
    # added tasks do not guarantee execution order or execution at all.

    def add_task(self, key: str, task: typing.Callable[[], None]) -> None:
        self.document_controller.add_task(key + str(id(self)), task)

    def clear_task(self, key: str) -> None:
        self.document_controller.clear_task(key + str(id(self)))

    def queue_task(self, task: typing.Callable[[], None]) -> None:
        self.document_controller.queue_task(task)

    def clear_queued_tasks(self) -> None:
        self.document_controller.clear_queued_tasks()

    def size_changed(self, width: int, height: int) -> None:
        pass

    def focus_changed(self, focused: bool) -> None:
        if focused:
            self.document_controller.request_refocus()

    def __physical_dpi_changed(self, physical_dpi: float) -> None:
        if self.dock_widget and self.dock_widget.widget:
            self.dock_widget.widget.redraw()

    def __str__(self) -> str:
        return self.display_name


class OutputPanel(Panel):

    __count = 0
    __old_stdout: typing.Optional[typing.TextIO] = None
    __old_stderr: typing.Optional[typing.TextIO] = None
    __stdout_listeners: typing.Dict[OutputPanel, typing.Callable[[str], None]] = dict()
    __stderr_listeners: typing.Dict[OutputPanel, typing.Callable[[str], None]] = dict()

    @classmethod
    def initialize(cls) -> None:
        """Configure standard output."""
        cls.__old_stdout = sys.stdout
        cls.__old_stderr = sys.stderr

        stdout_listeners = cls.__stdout_listeners

        class StdoutCatcher:
            def __init__(self, out: typing.TextIO) -> None:
                self.__out = out
            def write(self, stuff: typing.Any) -> None:
                for stdout_listener in stdout_listeners.values():
                    stdout_listener(stuff)
                self.__out.write(stuff)
            def flush(self) -> None:
                self.__out.flush()
            def getvalue(self) -> typing.Any:
                return typing.cast(typing.Any, self.__out).getvalue()
            @property
            def delegate(self) -> typing.Any:
                return typing.cast(typing.Any, self.__out).delegate

        # TODO: these casts are clearly wrong. spend more time in the future on this.
        sys.stdout = typing.cast(typing.TextIO, StdoutCatcher(cls.__old_stdout))
        sys.stderr = typing.cast(typing.TextIO, StdoutCatcher(cls.__old_stderr))

    @classmethod
    def deinitialize(cls) -> None:
        cls.__stdout_listeners = dict()
        cls.__stderr_listeners = dict()
        assert cls.__old_stdout
        assert cls.__old_stderr
        sys.stdout = cls.__old_stdout
        sys.stderr = cls.__old_stderr

    def __init__(self, document_controller: DocumentController.DocumentController, panel_id: str, properties: Persistence.PersistentDictType) -> None:
        super().__init__(document_controller, panel_id, "Output")
        properties["min-height"] = 180
        text_edit_widget = self.ui.create_text_edit_widget(properties)
        self.widget = text_edit_widget
        text_edit_widget.set_text_font(Panel.get_monospace_text_font())
        text_edit_widget.set_line_height_proportional(Panel.get_monospace_proportional_line_height())
        self.__lock = threading.RLock()
        self.__q = collections.deque()  # type: ignore  # Python 3.9+: collections.deque[str]

        def safe_emit() -> None:
            with self.__lock:
                while len(self.__q) > 0:
                    text_edit_widget.move_cursor_position("end")
                    text_edit_widget.append_text(self.__q.popleft())

        def queue_message(message: str) -> None:
            with self.__lock:
                self.__q.append(message.strip())
            if threading.current_thread().name == "MainThread":
                safe_emit()
            else:
                self.document_controller.queue_task(safe_emit)

        class OutputPanelHandler(logging.Handler):

            def __init__(self, queue_message_fn: typing.Callable[[str], None], records: typing.Sequence[logging.LogRecord]) -> None:
                super().__init__()
                self.queue_message_fn = queue_message_fn
                for record in records or list():
                    self.emit(record)

            def emit(self, record: logging.LogRecord) -> None:
                if record.levelno >= logging.INFO:
                    self.queue_message_fn(record.getMessage())

        self.__output_panel_handler = OutputPanelHandler(queue_message, Application.logging_handler.take_records())

        logging.getLogger().addHandler(self.__output_panel_handler)

        if OutputPanel.__count == 0:
            OutputPanel.initialize()

        OutputPanel.__stdout_listeners[self] = queue_message
        OutputPanel.__stderr_listeners[self] = queue_message

        OutputPanel.__count += 1

    def close(self) -> None:
        OutputPanel.__count -= 1
        OutputPanel.__stdout_listeners.pop(self)
        OutputPanel.__stderr_listeners.pop(self)

        if OutputPanel.__count == 0:
            OutputPanel.deinitialize()

        logging.getLogger().removeHandler(self.__output_panel_handler)
        super().close()


class CloseButtonCell(CanvasItem.Cell):
    def __init__(self) -> None:
        super().__init__()
        self.fill_style = "rgb(128, 128, 128)"
        self.fill_style_pressed = "rgb(64, 64, 64)"
        self.fill_style_disabled = "rgb(192, 192, 192)"
        self.border_style: str | None = None
        self.border_style_pressed: str | None = None
        self.border_style_disabled: str | None = None
        self.stroke_style = "#FFF"
        self.stroke_width = 3.0

    def _size_to_content(self, get_font_metrics_fn: typing.Callable[[str, str], UserInterface.FontMetrics]) -> Geometry.IntSize:
        return Geometry.IntSize(20, 20)

    def _paint_cell(self, drawing_context: DrawingContext.DrawingContext, canvas_bounds: Geometry.FloatRect, style: set[str]) -> None:
        if not Platform.is_macos():
            control_style = '#000000'
        else:
            control_style = '#808080'
        with drawing_context.saver():
            drawing_context.translate(canvas_bounds.left, canvas_bounds.top)
            drawing_context.begin_path()
            close_box_left = canvas_bounds.width - (20 - 7)
            close_box_right = canvas_bounds.width - (20 - 13)
            close_box_top = canvas_bounds.height // 2 - 3
            close_box_bottom = canvas_bounds.height // 2 + 3
            drawing_context.move_to(close_box_left, close_box_top)
            drawing_context.line_to(close_box_right, close_box_bottom)
            drawing_context.move_to(close_box_left, close_box_bottom)
            drawing_context.line_to(close_box_right, close_box_top)
            drawing_context.line_width = 1.5
            drawing_context.line_cap = "round"
            drawing_context.stroke_style = control_style
            drawing_context.stroke()


class CloseButtonCanvasItem(CanvasItem.CellCanvasItem):

    def __init__(self, ui_settings: UISettings.UISettings) -> None:
        super().__init__(CloseButtonCell())
        self.wants_mouse_events = True
        self.size_to_content(typing.cast(typing.Callable[[str, str], UserInterface.FontMetrics], ui_settings.get_font_metrics))


class HeaderBackgroundCanvasItemComposer(CanvasItem.BaseComposer):
    def __init__(self, canvas_item: CanvasItem.AbstractCanvasItem, layout_sizing: CanvasItem.Sizing,
                 cache: CanvasItem.ComposerCache, start_header_color: str, end_header_color: str, top_offset: int,
                 top_stroke_style: str | None, side_stroke_style: str | None, bottom_stroke_style: str | None) -> None:
        super().__init__(canvas_item, layout_sizing, cache)
        self.__start_header_color = start_header_color
        self.__end_header_color = end_header_color
        self.__top_offset = top_offset
        self.__top_stroke_style = top_stroke_style
        self.__side_stroke_style = side_stroke_style
        self.__bottom_stroke_style = bottom_stroke_style

    def _repaint(self, drawing_context: DrawingContext.DrawingContext, canvas_bounds: Geometry.IntRect, composer_cache: CanvasItem.ComposerCache) -> None:
        start_header_color = self.__start_header_color
        end_header_color = self.__end_header_color
        top_offset = self.__top_offset
        top_stroke_style = self.__top_stroke_style
        bottom_stroke_style = self.__bottom_stroke_style
        side_stroke_style = self.__side_stroke_style
        with drawing_context.saver():
            drawing_context.translate(canvas_bounds.left, canvas_bounds.top)
            with drawing_context.saver():
                drawing_context.begin_path()
                drawing_context.move_to(0, 1)
                drawing_context.line_to(0, canvas_bounds.height)
                drawing_context.line_to(canvas_bounds.width, canvas_bounds.height)
                drawing_context.line_to(canvas_bounds.width, 1)
                drawing_context.close_path()
                gradient = drawing_context.create_linear_gradient(canvas_bounds.width, canvas_bounds.height, 0, 0, 0, canvas_bounds.height)
                gradient.add_color_stop(0, start_header_color)
                gradient.add_color_stop(1, end_header_color)
                drawing_context.fill_style = gradient
                drawing_context.fill()

            with drawing_context.saver():
                drawing_context.begin_path()
                # line is adjust 1/2 pixel down to align to pixel boundary
                drawing_context.move_to(0, 0.5 + top_offset)
                drawing_context.line_to(canvas_bounds.width, 0.5 + top_offset)
                drawing_context.stroke_style = top_stroke_style
                drawing_context.stroke()

            with drawing_context.saver():
                drawing_context.begin_path()
                # line is adjust 1/2 pixel down to align to pixel boundary
                drawing_context.move_to(0, canvas_bounds.height-0.5)
                drawing_context.line_to(canvas_bounds.width, canvas_bounds.height-0.5)
                drawing_context.stroke_style = bottom_stroke_style
                drawing_context.stroke()

            if side_stroke_style:
                with drawing_context.saver():
                    drawing_context.begin_path()
                    # line is adjust 1/2 pixel down to align to pixel boundary
                    drawing_context.move_to(0.5, 1.5)
                    drawing_context.line_to(0.5, canvas_bounds.height - 0.5)
                    drawing_context.move_to(canvas_bounds.width - 0.5, 1.5)
                    drawing_context.line_to(canvas_bounds.width - 0.5, canvas_bounds.height - 0.5)
                    drawing_context.stroke_style = side_stroke_style
                    drawing_context.stroke()


class HeaderBackgroundCanvasItem(CanvasItem.AbstractCanvasItem):

    def __init__(self, height: int) -> None:
        super().__init__()
        self.__set_default_style()
        self.update_sizing(self.sizing.with_fixed_height(height))

    def __set_default_style(self) -> None:
        self.__side_stroke_style: str | None
        if not Platform.is_macos():
            self.__top_offset = 1
            self.__start_header_color = "#d9d9d9"
            self.__end_header_color = "#d9d9d9"
            self.__top_stroke_style = '#b8b8b8'
            self.__side_stroke_style = '#b8b8b8'
            self.__bottom_stroke_style = '#b8b8b8'
        else:
            self.__top_offset = 0
            self.__start_header_color = "#ededed"
            self.__end_header_color = "#cacaca"
            self.__top_stroke_style = '#ffffff'
            self.__bottom_stroke_style = '#b0b0b0'
            self.__side_stroke_style = None

    @property
    def start_header_color(self) -> str:
        return self.__start_header_color

    @start_header_color.setter
    def start_header_color(self, start_header_color: str) -> None:
        if self.__start_header_color != start_header_color:
            self.__start_header_color = start_header_color
            self.update()

    @property
    def end_header_color(self) -> str:
        return self.__end_header_color

    @end_header_color.setter
    def end_header_color(self, end_header_color: str) -> None:
        if self.__end_header_color != end_header_color:
            self.__end_header_color = end_header_color
            self.update()

    def reset_header_colors(self) -> None:
        self.__set_default_style()
        self.update()

    def _get_composer(self, composer_cache: CanvasItem.ComposerCache) -> CanvasItem.BaseComposer | None:
        return HeaderBackgroundCanvasItemComposer(self, self.sizing, composer_cache, self.__start_header_color,
                                                  self.__end_header_color, self.__top_offset, self.__top_stroke_style,
                                                  self.__side_stroke_style, self.__bottom_stroke_style)


class HeaderTitleCanvasItemComposer(CanvasItem.BaseComposer):
    def __init__(self, canvas_item: CanvasItem.AbstractCanvasItem, layout_sizing: CanvasItem.Sizing, cache: CanvasItem.ComposerCache, title: str, font: str, text_offset: int, ui_settings: UISettings.UISettings) -> None:
        super().__init__(canvas_item, layout_sizing, cache)
        self.__title = title
        self.__font = font
        self.__text_offset = text_offset
        self.__ui_settings = ui_settings

    def _repaint(self, drawing_context: DrawingContext.DrawingContext, canvas_bounds: Geometry.IntRect, composer_cache: CanvasItem.ComposerCache) -> None:
        font = self.__font
        title = self.__title
        text_offset = self.__text_offset
        canvas_bounds = canvas_bounds.intersect(Geometry.IntRect.from_tlbr(0, 0, canvas_bounds.bottom, canvas_bounds.right))
        title = self.__ui_settings.truncate_string_to_width(self.__font, title, canvas_bounds.width + 2, UISettings.TruncateModeType.MIDDLE)
        with drawing_context.saver():
            drawing_context.translate(canvas_bounds.left, canvas_bounds.top)
            drawing_context.clip_rect(0, 0, canvas_bounds.width, canvas_bounds.height)
            drawing_context.font = font
            drawing_context.text_align = 'left'
            drawing_context.text_baseline = 'bottom'
            drawing_context.fill_style = '#000'
            drawing_context.fill_text(title, 0, canvas_bounds.height - text_offset)


class HeaderTitleCanvasItem(CanvasItem.AbstractCanvasItem):

    def __init__(self, ui_settings: UISettings.UISettings, title: str, height: int, font: str, text_offset: int) -> None:
        super().__init__()
        self.__ui_settings = ui_settings
        self.__title = title
        self.__height = height
        self.__font = font
        self.__text_offset = text_offset
        self.update_sizing(self.sizing.with_fixed_height(height))

    def __str__(self) -> str:
        return self.__title

    @property
    def title(self) -> str:
        return self.__title

    @title.setter
    def title(self, title: str) -> None:
        if self.__title != title:
            self.__title = title
            font_metrics = self.__ui_settings.get_font_metrics(self.__font, self.__title)
            new_sizing = self.sizing
            new_sizing = new_sizing.with_preferred_width(font_metrics.width)
            self.update_sizing(new_sizing)
            self.update()

    def _get_composer(self, composer_cache: CanvasItem.ComposerCache) -> CanvasItem.BaseComposer | None:
        return HeaderTitleCanvasItemComposer(self, self.sizing, composer_cache, self.title, self.__font, self.__text_offset, self.__ui_settings)


class HeaderOverlayCanvasItem(CanvasItem.EmptyCanvasItem):

    def __init__(self, height: int) -> None:
        super().__init__()
        self.wants_mouse_events = True
        self.on_select_pressed: typing.Callable[[], None] | None = None
        self.on_drag_pressed: typing.Callable[[], None] | None = None
        self.on_context_menu_clicked: typing.Callable[[int, int, int, int], bool] | None = None
        self.on_double_clicked: typing.Callable[[int, int, UserInterface.KeyboardModifiers], bool] | None = None
        self.on_tool_tip: typing.Callable[[], str | None] | None = None
        self.__mouse_pressed_position: Geometry.IntPoint | None = None

    def mouse_entered(self) -> bool:
        if callable(self.on_tool_tip):
            self.tool_tip = self.on_tool_tip()
        return super().mouse_entered()

    def mouse_exited(self) -> bool:
        self.tool_tip = None
        return super().mouse_exited()

    def mouse_pressed(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        self.__mouse_pressed_position = Geometry.IntPoint(y=y, x=x)
        return True

    def mouse_double_clicked(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        if callable(self.on_double_clicked):
            return self.on_double_clicked(x, y, modifiers)
        return False

    def mouse_position_changed(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        pt = Geometry.FloatPoint(y=y, x=x)
        mouse_pressed_pos = self.__mouse_pressed_position
        if mouse_pressed_pos and Geometry.distance(mouse_pressed_pos.to_float_point(), pt) > 12:
            on_drag_pressed = self.on_drag_pressed
            if callable(on_drag_pressed):
                self.__mouse_pressed_position = None
                on_drag_pressed()
                return True
        return False

    def mouse_released(self, x: int, y: int, modifiers: UserInterface.KeyboardModifiers) -> bool:
        canvas_size = self.canvas_size
        if self.__mouse_pressed_position is not None:
            on_select_pressed = self.on_select_pressed
            if callable(on_select_pressed):
                on_select_pressed()
        self.__mouse_pressed_position = None
        return True

    def context_menu_event(self, x: int, y: int, gx: int, gy: int) -> bool:
        if callable(self.on_context_menu_clicked):
            return self.on_context_menu_clicked(x, y, gx, gy)
        return False


class HeaderCanvasItem(CanvasItem.CanvasItemComposition):

    def __init__(self, ui_settings: UISettings.UISettings, title: str | None = None, display_close_control: bool = False) -> None:
        super().__init__()
        self.__ui_settings = ui_settings
        if not Platform.is_macos():
            self.__font = 'normal 11px system serif'
            self.__text_offset = 4
        else:
            self.__font = 'normal 10pt system serif'
            self.__text_offset = 7
        height = self.__ui_settings.get_font_metrics(self.__font, "abc").height + 3 + self.__text_offset
        self.__title = title or str()
        self.__title_canvas_item = HeaderTitleCanvasItem(self.__ui_settings, self.__title, height, self.__font, self.__text_offset)
        self.__overlay_canvas_item = HeaderOverlayCanvasItem(height)
        self.__header_background_canvas_item = HeaderBackgroundCanvasItem(height)
        self.__close_button_canvas_item = CloseButtonCanvasItem(self.__ui_settings) if display_close_control else None
        title_row = CanvasItem.CanvasItemComposition()
        title_row.layout = CanvasItem.CanvasItemRowLayout()
        title_row.add_stretch()
        title_row.add_canvas_item(self.__title_canvas_item)
        title_row.add_stretch()
        title_row.update_sizing(title_row.sizing.with_minimum_width(0))
        header_row = CanvasItem.CanvasItemComposition()
        header_row.layout = CanvasItem.CanvasItemRowLayout()
        header_row.add_spacing(6)
        header_overlay = CanvasItem.CanvasItemComposition()
        header_overlay.add_canvas_item(title_row)
        header_overlay.add_canvas_item(self.__overlay_canvas_item)
        header_row.add_canvas_item(header_overlay)
        if self.__close_button_canvas_item:
            header_row.add_canvas_item(self.__close_button_canvas_item)
        self.add_canvas_item(self.__header_background_canvas_item)
        self.add_canvas_item(header_row)
        self.update_sizing(self.sizing.with_fixed_height(height))

    @property
    def header_height(self) -> int:
        return self.__ui_settings.get_font_metrics(self.__font, "abc").height + 3 + self.__text_offset

    @property
    def title(self) -> str:
        return self.__title_canvas_item.title

    @title.setter
    def title(self, title: str) -> None:
        self.__title_canvas_item.title = title

    @property
    def start_header_color(self) -> str:
        return self.__header_background_canvas_item.start_header_color

    @start_header_color.setter
    def start_header_color(self, start_header_color: str) -> None:
        self.__header_background_canvas_item.start_header_color = start_header_color

    @property
    def end_header_color(self) -> str:
        return self.__header_background_canvas_item.end_header_color

    @end_header_color.setter
    def end_header_color(self, end_header_color: str) -> None:
        self.__header_background_canvas_item.end_header_color = end_header_color

    def reset_header_colors(self) -> None:
        self.__header_background_canvas_item.reset_header_colors()

    @property
    def on_select_pressed(self) -> typing.Callable[[], None] | None:
        return self.__overlay_canvas_item.on_select_pressed

    @on_select_pressed.setter
    def on_select_pressed(self, value: typing.Callable[[], None] | None) -> None:
        self.__overlay_canvas_item.on_select_pressed = value

    @property
    def on_drag_pressed(self) -> typing.Callable[[], None] | None:
        return self.__overlay_canvas_item.on_drag_pressed

    @on_drag_pressed.setter
    def on_drag_pressed(self, value: typing.Callable[[], None] | None) -> None:
        self.__overlay_canvas_item.on_drag_pressed = value

    @property
    def on_context_menu_clicked(self) -> typing.Callable[[int, int, int, int], bool] | None:
        return self.__overlay_canvas_item.on_context_menu_clicked

    @on_context_menu_clicked.setter
    def on_context_menu_clicked(self, value: typing.Callable[[int, int, int, int], bool] | None) -> None:
        self.__overlay_canvas_item.on_context_menu_clicked = value

    @property
    def on_double_clicked(self) -> typing.Callable[[int, int, UserInterface.KeyboardModifiers], bool] | None:
        return self.__overlay_canvas_item.on_double_clicked

    @on_double_clicked.setter
    def on_double_clicked(self, value: typing.Callable[[int, int, UserInterface.KeyboardModifiers], bool] | None) -> None:
        self.__overlay_canvas_item.on_double_clicked = value

    @property
    def on_tool_tip(self) -> typing.Callable[[], str | None] | None:
        return self.__overlay_canvas_item.on_tool_tip

    @on_tool_tip.setter
    def on_tool_tip(self, value: typing.Callable[[], str | None] | None) -> None:
        self.__overlay_canvas_item.on_tool_tip = value

    @property
    def on_close_clicked(self) -> typing.Callable[[], None] | None:
        return self.__close_button_canvas_item.on_clicked if self.__close_button_canvas_item else None

    @on_close_clicked.setter
    def on_close_clicked(self, value: typing.Callable[[], None] | None) -> None:
        if self.__close_button_canvas_item:
            self.__close_button_canvas_item.on_clicked = value

    # for testing
    def simulate_click(self, p: Geometry.IntPointTuple, modifiers: typing.Optional[UserInterface.KeyboardModifiers] = None) -> None:
        self.__overlay_canvas_item.simulate_click(p, modifiers)


class PanelSectionFactory(typing.Protocol):
    COMPONENT_TYPE: str = "panel_section"

    panel_section_ids: typing.Set[str]

    def make_panel_section(self, panel_section_id: str, d: Persistence.PersistentDictType, document_controller: DocumentController.DocumentController, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        ...


class SectionPanelSection(Declarative.Handler):
    def __init__(self, panel_section_handler: Declarative.HandlerLike) -> None:
        super().__init__()
        self.panel_section_handler = panel_section_handler
        self.is_expanded = True
        panel_section_title = getattr(panel_section_handler, "title")
        u = Declarative.DeclarativeUI()
        self.ui_view = u.create_section(u.create_component_instance(identifier="panel_section_handler"), title=panel_section_title, expanded=f"@binding(is_expanded)")

    def create_handler(self, component_id: str, container: typing.Any = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        return self.panel_section_handler


class SectionPanelHandler(Declarative.Handler):
    def __init__(self, panel_section_handlers: typing.Sequence[Declarative.HandlerLike], properties: Persistence.PersistentDictType) -> None:
        super().__init__()
        self.panel_section_handlers = list(panel_section_handlers)
        u = Declarative.DeclarativeUI()
        self.ui_view = u.create_column(u.create_column(items="panel_section_handlers", item_component_id="section"), u.create_stretch(), **properties)

    def create_handler(self, component_id: str, container: typing.Any = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        return SectionPanelSection(item)


class SectionPanel(Panel):
    def __init__(self, panel_d: Persistence.PersistentDictType, document_controller: DocumentController.DocumentController, panel_id: str, properties: Persistence.PersistentDictType) -> None:
        super().__init__(document_controller, panel_id, "section-panel")
        panel_section_handlers: typing.List[Declarative.HandlerLike] = list()
        for panel_section_d in panel_d.get("panel_sections", list()):
            panel_section_id = panel_section_d.get("panel_section_id", str())
            for panel_section_factory in typing.cast(typing.Set[PanelSectionFactory], Registry.get_components_by_type(PanelSectionFactory.COMPONENT_TYPE)):
                panel_section_handler = panel_section_factory.make_panel_section(panel_section_id, panel_section_d, document_controller)
                if panel_section_handler:
                    panel_section_handlers.append(panel_section_handler)
                    break
        assert panel_section_handlers
        self.widget = Declarative.DeclarativeWidget(document_controller.ui, document_controller.event_loop, SectionPanelHandler(panel_section_handlers, properties))


PanelCreateFn = typing.Callable[["DocumentController.DocumentController", str, typing.Optional["Persistence.PersistentDictType"]], Panel]


@dataclasses.dataclass
class PanelTuple:
    panel_create_fn: PanelCreateFn
    panel_id: str
    name: str
    positions: typing.List[str]
    position: str
    properties: typing.Optional[Persistence.PersistentDictType]


class PanelManager(metaclass=Utility.Singleton):
    def __init__(self) -> None:
        self.__panel_tuples: typing.Dict[str, PanelTuple] = dict()

    def register_panel(self, panel_class: PanelCreateFn, panel_id: str, name: str,
                       positions: typing.Sequence[str], position: str,
                       properties: typing.Optional[Persistence.PersistentDictType] = None) -> None:
        self.__panel_tuples[panel_id] = PanelTuple(panel_class, panel_id, name, list(positions), position, properties)

    def unregister_panel(self, panel_id: str) -> None:
        del self.__panel_tuples[panel_id]

    def create_panel_content(self, document_controller: DocumentController.DocumentController, panel_id: str,
                             title: str, positions: typing.Sequence[str], position: str,
                             properties: typing.Optional[Persistence.PersistentDictType]) -> typing.Optional[Panel]:
        if panel_id in self.__panel_tuples:
            panel_tuple = self.__panel_tuples[panel_id]
            try:
                properties = properties if properties else {}
                panel: Panel = panel_tuple.panel_create_fn(document_controller, panel_id, properties)
                panel.create_dock_widget(title, positions, position)
                return panel
            except Exception as e:
                import traceback
                print("Exception creating panel '" + panel_id + "': " + str(e))
                traceback.print_exc()
                traceback.print_stack()
        return None

    def get_panel_info(self, panel_id: str) -> typing.Tuple[str, typing.Sequence[str], str, typing.Optional[Persistence.PersistentDictType]]:
        panel_tuple = self.__panel_tuples[panel_id]
        return panel_tuple.name, panel_tuple.positions, panel_tuple.position, panel_tuple.properties

    @property
    def panel_ids(self) -> typing.List[str]:
        return list(self.__panel_tuples.keys())

    def load(self, d: typing.Sequence[Persistence.PersistentDictType]) -> None:
        for panel_d in d:
            if panel_d.get("version", 0) == 1:
                section_panel_id = panel_d["panel_id"]
                section_panel_title = panel_d["title"]
                is_panel_valid = False
                for panel_section_d in panel_d.get("panel_sections", list()):
                    panel_section_id = panel_section_d.get("panel_section_id", str())
                    if self.is_panel_section_valid(panel_section_id):
                        is_panel_valid = True
                        break
                if is_panel_valid:
                    def create_panel(panel_d: Persistence.PersistentDictType,
                                     document_controller: DocumentController.DocumentController, panel_id: str,
                                     properties: Persistence.PersistentDictType) -> Panel:
                        return SectionPanel(panel_d, document_controller, panel_id, properties)

                    self.register_panel(typing.cast(PanelCreateFn, functools.partial(create_panel, panel_d)),
                                        section_panel_id, section_panel_title, ["left", "right"], "right",
                                        {"min_width": 320})

    def is_panel_section_valid(self, panel_section_id: str) -> bool:
        for panel_section_factory in typing.cast(typing.Set[PanelSectionFactory], Registry.get_components_by_type(PanelSectionFactory.COMPONENT_TYPE)):
            if panel_section_id in panel_section_factory.panel_section_ids:
                return True
        return False


# useful utility class

class MappedListModelLike(typing.Protocol):
    item_inserted_event: Event.Event
    item_removed_event: Event.Event
    begin_changes_event: Event.Event
    end_changes_event: Event.Event

    @property
    def items(self) -> typing.Sequence[typing.Any]:
        raise NotImplementedError()


class ThreadSafeListModel(MappedListModelLike):
    def __init__(self, list_model: MappedListModelLike, event_loop: asyncio.AbstractEventLoop) -> None:
        self.__items = list(list_model.items)
        self.__event_loop = event_loop

        self.__list_model = list_model
        self.item_inserted_event = Event.Event()
        self.item_removed_event = Event.Event()
        self.begin_changes_event = Event.Event()
        self.end_changes_event = Event.Event()

        self.__item_inserted_listener = self.__list_model.item_inserted_event.listen(ReferenceCounting.weak_partial(ThreadSafeListModel.__list_model_item_inserted, self))
        self.__item_removed_listener = self.__list_model.item_removed_event.listen(ReferenceCounting.weak_partial(ThreadSafeListModel.__list_model_item_removed, self))
        self.__begin_changes_listener = self.__list_model.begin_changes_event.listen(ReferenceCounting.weak_partial(ThreadSafeListModel.__begin_changes, self))
        self.__end_changes_listener = self.__list_model.end_changes_event.listen(ReferenceCounting.weak_partial(ThreadSafeListModel.__end_changes, self))

    def __list_model_item_inserted(self, key: str, item: typing.Any, before_index: int) -> None:
        if threading.current_thread() != threading.main_thread():
            self.__event_loop.call_soon_threadsafe(self.__list_model_item_inserted, key, item, before_index)
        else:
            self.__items.insert(before_index, item)
            self.item_inserted_event.fire(key, item, before_index)

    def __list_model_item_removed(self, key: str, item: typing.Any, index: int) -> None:
        if threading.current_thread() != threading.main_thread():
            self.__event_loop.call_soon_threadsafe(self.__list_model_item_removed, key, item, index)
        else:
            del self.__items[index]
            self.item_removed_event.fire(key, item, index)

    def __begin_changes(self, key: str) -> None:
        self.begin_changes_event.fire(key)

    def __end_changes(self, key: str) -> None:
        self.end_changes_event.fire(key)

    @property
    def items(self) -> typing.Sequence[typing.Any]:
        return list(self.__items)



# standard libraries
import contextlib
import typing
import unittest

# third party libraries
import numpy

# local libraries
from nion.swift import Application
from nion.swift import Facade
from nion.swift import RecorderPanel
from nion.swift.model import DataItem
from nion.swift.test import TestContext
from nion.ui import TestUI


Facade.initialize()

class TestRecorderPanelClass(unittest.TestCase):

    def setUp(self):
        TestContext.begin_leaks()
        self._test_setup = TestContext.TestSetup()

    def tearDown(self):
        self._test_setup = typing.cast(typing.Any, None)
        TestContext.end_leaks(self)

    def test_recorder_records_live_data(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.ones((8, 8)))
            document_model.append_data_item(data_item)
            recorder = RecorderPanel.Recorder(document_controller, data_item)
            with contextlib.closing(recorder):
                with document_model.data_item_live(data_item):
                    count = 4
                    recorder.start_recording(10, 1, count)
                    for i in range(count):
                        recorder.continue_recording(10 + i + 0.25)
                        recorder.continue_recording(10 + i + 0.75)
                        data_item.set_data(data_item.data + 1)
            self.assertEqual(2, len(document_model.data_items))
            recorded_data_item = document_model.data_items[1]
            self.assertTrue(recorded_data_item.xdata.is_sequence)
            self.assertEqual((4, 8, 8), recorded_data_item.xdata.dimensional_shape)

    def test_recorder_puts_recorded_data_item_under_transaction(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.ones((8, 8)))
            document_model.append_data_item(data_item)
            recorder = RecorderPanel.Recorder(document_controller, data_item)
            with contextlib.closing(recorder):
                with document_model.data_item_live(data_item):
                    count = 4
                    recorder.start_recording(10, 1, count)
                    for i in range(count):
                        recorder.continue_recording(10 + i + 0.25)
                        data_item.set_data(data_item.data + 1)
                        if i != count - 1:  # last iteration it should fall out of transaction state since it is finished
                            self.assertTrue(document_model.data_items[1].in_transaction_state)
                        else:
                            self.assertFalse(document_model.data_items[1].in_transaction_state)
            self.assertFalse(document_model.data_items[1].in_transaction_state)

    def test_recorder_state_is_reported_properly(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            data_item = DataItem.DataItem(numpy.ones((8, 8)))
            document_model.append_data_item(data_item)
            recorder_state_ref = ["unknown"]
            def recorder_state_changed(recorder_state): recorder_state_ref[0] = recorder_state
            recorder = RecorderPanel.Recorder(document_controller, data_item)
            recorder.on_recording_state_changed = recorder_state_changed
            with contextlib.closing(recorder):
                with document_model.data_item_live(data_item):
                    count = 4
                    recorder.start_recording(10, 1, count)
                    for i in range(count):
                        recorder.continue_recording(10 + i + 0.25)
                        data_item.set_data(data_item.data + 1)
                        if i != count - 1:  # last iteration it should fall out of transaction state since it is finished
                            self.assertEqual(recorder_state_ref[0], "recording")
                        else:
                            self.assertEqual(recorder_state_ref[0], "stopped")
                self.assertEqual(recorder_state_ref[0], "stopped")



"""
Provide tools for a general entity browser.

An entity is defined by a schema and this set of UI elements facilitates a master-detail hierarchy to browse
entities (aka items/objects) that adhere to that schema.
"""

from __future__ import annotations

# standard libraries
import asyncio
import dataclasses
import gettext
import operator
import threading
import typing

# third party libraries
# None

# local libraries
from nion.swift.model import Schema
from nion.ui import Declarative
from nion.ui import UserInterface
from nion.ui import Widgets
from nion.ui import Window
from nion.utils import Converter
from nion.utils import Event
from nion.utils import ListModel
from nion.utils import Model
from nion.utils import Observable
from nion.utils import Registry
from nion.utils import ReferenceCounting

_ = gettext.gettext

T = typing.TypeVar('T')

INDENT = 8
LABEL_WIDTH = 100


class ReferenceHandlerContext(typing.Protocol):
    def open_project_item(self, item: typing.Any) -> None: ...


@dataclasses.dataclass
class Context:
    values: typing.Dict[str, typing.Any] = dataclasses.field(default_factory=dict)


class ReferenceHandler(Declarative.Handler):
    def __init__(self, context: Context, label: str, item: typing.Optional[typing.Any]) -> None:
        super().__init__()
        self.context = context
        self.label = label
        self.item = item
        self.ui_view = self.__make_ui()

    def __make_ui(self) -> Declarative.UIDescriptionResult:
        u = Declarative.DeclarativeUI()
        if self.item:
            label = u.create_label(text=self.item.__class__.__name__)
            link = u.create_push_button(text="\N{RIGHTWARDS BLACK ARROW}", on_clicked="handle_link",
                                        border_color="transparent", background_color="rgba(0,0,0,0.0)", style="minimal",
                                        size_policy_horizontal="maximum")
            label_section = u.create_row(label, link)
        else:
            label_section = u.create_label(text=_("None [reference]"))
        return u.create_row(
            u.create_label(text=self.label, tool_tip=self.label, width=LABEL_WIDTH),
            label_section,
            u.create_stretch(),
            spacing=12)

    def handle_link(self, item: Declarative.UIWidget) -> None:
        if self.item:
            reference_handler = typing.cast(typing.Optional[ReferenceHandlerContext], self.context.values.get("reference_handler", None))
            if reference_handler:
                reference_handler.open_project_item(self.item)


class ItemToStringConverter(Converter.ConverterLike[typing.Any, str]):
    def convert(self, value: typing.Optional[typing.Any]) -> typing.Optional[str]:
        return f"{str(value)}" if value else _("<empty>")

    def convert_back(self, formatted_value: typing.Optional[str]) -> typing.Optional[typing.Any]:
        return str()


class EntityPropertyHandler(Declarative.Handler):
    def __init__(self, name: str, value_type: Schema.PropertyType, value_model: Model.PropertyModel[typing.Any]) -> None:
        super().__init__()
        self.value_model = value_model
        self.ui_view = self._make_ui(name, value_type)
        self.int_converter = Converter.IntegerToStringConverter()
        self.float_converter = Converter.FloatToStringConverter()
        self.date_converter = Converter.DatetimeToStringConverter(is_local=True, format="%Y-%m-%d %H:%M:%S %Z")
        self.uuid_converter = Converter.UuidToStringConverter()
        self.str_converter = ItemToStringConverter()

    def _make_ui(self, name: str, value_type: Schema.PropertyType) -> Declarative.UIDescriptionResult:
        u = Declarative.DeclarativeUI()
        if value_type.type == Schema.STRING:
            return u.create_row(u.create_label(text=name, tool_tip=name, width=LABEL_WIDTH),
                                u.create_label(text=f"@binding(value_model.value, converter=str_converter)"),
                                u.create_stretch(), spacing=12)
        elif value_type.type == Schema.BOOLEAN:
            return u.create_row(u.create_label(text=name, tool_tip=name, width=LABEL_WIDTH),
                                u.create_check_box(text=f"@binding(value_model.value)"),
                                u.create_stretch(), spacing=12)
        elif value_type.type == Schema.INT:
            return u.create_row(u.create_label(text=name, tool_tip=name, width=LABEL_WIDTH),
                                u.create_label(text=f"@binding(value_model.value, converter=int_converter)"),
                                u.create_stretch(), spacing=12)
        elif value_type.type == Schema.FLOAT:
            return u.create_row(u.create_label(text=name, tool_tip=name, width=LABEL_WIDTH),
                                u.create_label(text=f"@binding(value_model.value, converter=float_converter)"),
                                u.create_stretch(), spacing=12)
        elif value_type.type == Schema.TIMESTAMP:
            return u.create_row(u.create_label(text=name, tool_tip=name, width=LABEL_WIDTH),
                                u.create_label(text=f"@binding(value_model.value, converter=date_converter)"),
                                u.create_stretch(), spacing=12)
        elif value_type.type == Schema.UUID:
            return u.create_row(u.create_label(text=name, tool_tip=name, width=LABEL_WIDTH),
                                u.create_label(text=f"@binding(value_model.value, converter=uuid_converter)"),
                                u.create_stretch(), spacing=12)
        elif value_type.type in (Schema.DICT, Schema.LIST, Schema.SET):
            return u.create_row(u.create_label(text=name, tool_tip=name, width=LABEL_WIDTH),
                                u.create_row(
                                    u.create_label(text=f"{value_type.type}: "),
                                    u.create_label(text=f"@binding(value_model.value, converter=str_converter)"),
                                ),
                                u.create_stretch(),
                                spacing=12)
        NOT_DISPLAYED = _("not displayed")
        return u.create_row(u.create_label(text=name, tool_tip=name, width=LABEL_WIDTH),
                            u.create_label(text=f"{value_type.type} ({NOT_DISPLAYED})"),
                            u.create_stretch(),
                            spacing=12)


class EntityTupleModel(Observable.Observable):
    # takes a property model and generates item inserted/removed events when the value changes.
    # observers can treat this as a dynamic list with 'items' key.

    def __init__(self, value_model: Model.PropertyModel[typing.Any]) -> None:
        super().__init__()
        self.items: typing.List[typing.Any] = list()

        def property_changed(tuple_model: EntityTupleModel, property_name: str) -> None:
            # check if changed property matches property name for this object
            if property_name == "value":
                while tuple_model.items:
                    item = tuple_model.items.pop()
                    tuple_model.notify_remove_item("items", item, len(tuple_model.items))
                items = typing.cast(typing.List[typing.Any], value_model.value)
                for index, item in enumerate(items or tuple()):
                    tuple_model.items.append((index, item))
                    tuple_model.notify_insert_item("items", tuple_model.items[-1], len(tuple_model.items) - 1)

        self.__listener = value_model.property_changed_event.listen(ReferenceCounting.weak_partial(property_changed, self))

        property_changed(self, "value")


class EntityTupleHandler(Declarative.Handler):
    def __init__(self, context: Context, name: str, value_type: Schema.FieldType, value_model: Model.PropertyModel[typing.Any]) -> None:
        super().__init__()
        self.context = context
        self.value_model = value_model
        self.value_type = value_type
        self.tuple_model = EntityTupleModel(value_model)
        u = Declarative.DeclarativeUI()
        # the items in tuple_model (tuples an index and a value of value_type) will get passed to create_handler in the
        # item parameter. this is accomplished by passing items to create_column below. the item_component_id is
        # used to match this column with the request for a handler.
        self.ui_view = u.create_column(
            u.create_row(u.create_label(text=name), u.create_stretch()),
            u.create_row(u.create_spacing(INDENT), u.create_column(items="tuple_model.items", item_component_id="entity_field", spacing=8)),
            spacing = 8,
        )

    def create_handler(self, component_id: str, container: typing.Optional[ListModel.ListModel[typing.Any]] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        # item is a tuple of type index, self.value_type
        if component_id == "entity_field" and item is not None:
            index, item_ = item
            return make_field_handler(self.context, str(index), self.value_type, Model.PropertyModel(item_))
        return None


class EntityArrayHandler(Declarative.Handler):
    def __init__(self, context: Context, name: str, value_type: Schema.FieldType, value_model: Model.PropertyModel[typing.Any]) -> None:
        super().__init__()
        self.context = context
        self.value_model = value_model
        self.value_type = value_type
        self.tuple_model = EntityTupleModel(value_model)
        u = Declarative.DeclarativeUI()
        # the items in tuple_model (tuples an index and a value of value_type) will get passed to create_handler in the
        # item parameter. this is accomplished by passing items to create_column below. the item_component_id is
        # used to match this column with the request for a handler.
        self.ui_view = u.create_column(
            u.create_row(u.create_label(text=name), u.create_stretch()),
            u.create_row(u.create_spacing(INDENT), u.create_column(items="tuple_model.items", item_component_id="entity_field", spacing=8)),
            spacing = 8,
        )

    def create_handler(self, component_id: str, container: typing.Optional[ListModel.ListModel[typing.Any]] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        # item is a tuple of type index, self.value_type
        if component_id == "entity_field" and item is not None:
            index, item_ = item
            return make_field_handler(self.context, str(index), self.value_type, Model.PropertyModel(item_), display_component_type=True)
        return None


class EntityValueIndexModel(Model.PropertyModel[T], typing.Generic[T]):
    # NOTE: read only

    def __init__(self, value_model: Model.PropertyModel[T], index: int) -> None:
        assert value_model.value is not None
        super().__init__(getattr(value_model, "value")[index])
        self.__value_model = value_model

        def property_changed(property_model: EntityValueIndexModel[T], value_model: Model.PropertyModel[T], property_name: str) -> None:
            # check if changed property matches property name for this object
            if property_name == "value":
                property_model.value = getattr(value_model, property_name)[index]

        self.__listener = self.__value_model.property_changed_event.listen(ReferenceCounting.weak_partial(property_changed, self, self.__value_model))


class EntityFieldsHandler(Declarative.Handler):
    def __init__(self, context: Context, name: str, field_list: typing.Sequence[typing.Tuple[str, Schema.FieldType, Model.PropertyModel[typing.Any]]], *, component_type: typing.Optional[str] = None) -> None:
        super().__init__()
        self.context = context
        self.field_list = list(field_list)
        u = Declarative.DeclarativeUI()
        # the items in field_list (field-name/value-type/property-model) will get passed to create_handler in the
        # item parameter. this is accomplished by passing items to create_column below. the item_component_id is
        # used to match this column with the request for a handler.
        component_type_row = u.create_row(u.create_label(text=_("type"), width=LABEL_WIDTH), u.create_label(text=component_type), u.create_stretch(), spacing=12)
        self.ui_view = u.create_column(
            u.create_row(u.create_label(text=name), u.create_stretch()),
            u.create_row(
                u.create_spacing(INDENT),
                u.create_column(
                    *([component_type_row] if component_type else []),
                    u.create_row(u.create_column(items="field_list", item_component_id="entity_field", spacing=8)),
                    spacing=8
                ),
            ),
            spacing = 8,
        )

    def create_handler(self, component_id: str, container: typing.Optional[typing.Any] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        # item is a tuple of field name, value type, and value model. it comes from the field list passed during init.
        if component_id == "entity_field" and item is not None:
            field_name, value_type, value_model = typing.cast(typing.Tuple[str, Schema.FieldType, Model.PropertyModel[typing.Any]], item)
            return make_field_handler(self.context, field_name, value_type, value_model)
        return None


class MaybePropertyChangedPropertyModel(Model.PropertyModel[typing.Any]):
    """Observes a property on another item and makes it a standard property model.

    When the observed property changes, update this value.

    When this value changes, update the observed property.
    """

    def __init__(self, observable: Observable.Observable, property_name: str) -> None:
        super().__init__(getattr(observable, property_name, None))
        self.__observable = observable
        self.__property_name = property_name

        def property_changed(property_model: MaybePropertyChangedPropertyModel, observable: Observable.Observable, property_name: str, property_name_: str) -> None:
            # check if changed property matches property name for this object
            if property_name_ == property_name:
                property_model.value = getattr(observable, property_name)

        if hasattr(self.__observable, "property_changed_event"):
            self.__listener = self.__observable.property_changed_event.listen(ReferenceCounting.weak_partial(property_changed, self, observable, property_name))

    def _set_value(self, value: typing.Optional[T]) -> None:
        super()._set_value(value)
        # set the property on the observed object. this will trigger a property changed, but will be ignored since
        # the value doesn't change.
        try:
            setattr(self.__observable, self.__property_name, value)
        except AttributeError as e:
            # as a convenience for handling dependent read-only fields, such as data_shape on a data item, ignore
            # attributes that can't be set. it's not clear if this is the right solution...
            pass


class DummyHandler(Declarative.HandlerLike):
    def __init__(self, field_name: str, text: str) -> None:
        u = Declarative.DeclarativeUI()
        self.ui_view = u.create_row(u.create_label(text=field_name, tool_tip=field_name, width=LABEL_WIDTH), u.create_label(text=text), u.create_stretch(), spacing=12)

    def close(self) -> None:
        pass


class HasFieldTypeMap(typing.Protocol):
    @property
    def _field_type_map(self) -> typing.Mapping[str, Schema.FieldType]: raise NotImplementedError()


def make_record_handler(context: Context, item: Observable.Observable, name: str, entity_type: HasFieldTypeMap, *, display_component_type: bool = False) -> EntityFieldsHandler:
    field_list = [(n, t, MaybePropertyChangedPropertyModel(item, n)) for n, t in entity_type._field_type_map.items()]
    return EntityFieldsHandler(context, name, field_list, component_type=getattr(entity_type, "entity_id", None))


def make_field_handler(context: Context, field_name: str, value_type: Schema.FieldType, value_model: Model.PropertyModel[typing.Any], *, display_component_type: bool = False) -> typing.Optional[Declarative.HandlerLike]:
    # when item type is property type, create an entity property handler, passing the value type and property model directly.
    if isinstance(value_type, Schema.PropertyType):
        return EntityPropertyHandler(field_name, value_type, value_model)

    # when item type is tuple type, create an entity tuple handler, passing the value type of the tuple and the property model directly.
    # the tuple handler will watch for changes to the property model and update its internal list accordingly.
    if isinstance(value_type, Schema.TupleType):
        return EntityTupleHandler(context, field_name, value_type.type, value_model)

    # when item type is fixed tuple type, make a list of field-name/value-type/property-model tuples, creating
    # entity value index models to access the individual fields, and recursively create another entity fields
    # handler with the list.
    if isinstance(value_type, Schema.FixedTupleType):
        if value_model.value is not None:
            field_list = [(str(i), t, EntityValueIndexModel(value_model, i)) for i, t in enumerate(value_type.types)]
            return EntityFieldsHandler(context, field_name, field_list)
        else:
            return DummyHandler(field_name, "NONE [fixed-tuple]")

    # when item type is record type, use make_record_handler to where the item is the value of the value_model passed
    # to this function.
    if isinstance(value_type, Schema.RecordType):
        if value_model.value is not None:
            return make_record_handler(context, value_model.value, field_name, value_type)
        else:
            return DummyHandler(field_name, "NONE [record]")

    # when item type is array type, create an entity array handler, passing the value type of the array and the property model directly.
    # the array handler will watch for changes to either the property model or the list value of the property model and update its
    # internal list accordingly.
    if isinstance(value_type, Schema.ArrayType):
        return EntityArrayHandler(context, field_name, value_type.type, value_model)

    # map

    # reference is only used if allowed in the context.
    if isinstance(value_type, Schema.ReferenceType):
        if context.values.get("do_references", False):
            return ReferenceHandler(context, field_name, value_model.value)

    # when item type is component type, use make_record_handler to where the item is the value of the value_model passed
    # to this function. get the entity type from the component type.
    if isinstance(value_type, Schema.ComponentType):
        if value_model.value is not None:
            entity_type = Schema.get_entity_type(value_type.entity_id)
            # hack to handle subclasses; look at the value itself and figure out the type if possible.
            if value_model.value and hasattr(value_model.value, "type") and value_model.value.type != value_type.entity_id:
                entity_type = Schema.entity_types.get(value_model.value.type, entity_type)
            assert entity_type
            return make_record_handler(context, value_model.value, field_name, entity_type, display_component_type=display_component_type)
        else:
            return DummyHandler(field_name, "NONE [component]")

    # fall through to a dummy handler.
    return DummyHandler(field_name, str(value_type))


class ItemInspectorHandlerFactory(typing.Protocol):
    def make_sections(self, context: Context, item: typing.Any) -> typing.Sequence[typing.Tuple[str, Declarative.HandlerLike]]: ...


class ItemPageHandler(Declarative.Handler):
    def __init__(self,
                 context: Context,
                 item: typing.Any,
                 entity_type: typing.Optional[Schema.EntityType],
                 title: str) -> None:
        super().__init__()
        self.context = context
        self.item = item
        self.__title = title
        self.__entity_type = entity_type
        self.__handlers: typing.List[Declarative.HandlerLike] = list()
        self.uuid_converter = Converter.UuidToStringConverter()
        self.date_converter = Converter.DatetimeToStringConverter(is_local=True, format="%Y-%m-%d %H:%M:%S %Z")
        self.ui_view = self._make_ui()

    def _make_ui(self) -> Declarative.UIDescriptionResult:
        tabs: typing.List[Declarative.UIDescription] = list()
        item_inspector_sections: typing.List[typing.Tuple[str, Declarative.HandlerLike]] = list()
        for component in Registry.get_components_by_type("item-inspector-handler-factory"):
            item_inspector_handler_factory = typing.cast(ItemInspectorHandlerFactory, component)
            item_inspector_sections.extend(item_inspector_handler_factory.make_sections(self.context, self.item))
        u = Declarative.DeclarativeUI()
        for index, (title, handler) in enumerate(item_inspector_sections):
            handler_index = len(self.__handlers)
            self.__handlers.append(handler)
            tabs.append(u.create_tab(title, u.create_component_instance(f"handler_{handler_index}")))
        if self.__entity_type:
            entity_component = u.create_scroll_area(u.create_column(u.create_component_instance("entity_component"), u.create_stretch()))
            tabs.append(u.create_tab(_("Details"), entity_component))
        return u.create_tabs(*tabs, style="minimal")

    def create_handler(self, component_id: str, container: typing.Optional[typing.Any] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        if component_id == "entity_component":
            # hack to handle subclasses; look at the value itself and figure out the type if possible.
            entity_type = self.__entity_type
            assert entity_type
            if self.item and hasattr(self.item, "type") and self.item.type in [t.entity_id for t in entity_type.subclasses]:
                entity_type = Schema.entity_types.get(self.item.type, entity_type)
                assert entity_type
            return make_record_handler(self.context, self.item, entity_type.entity_id, entity_type)
        for handler_index, handler in enumerate(self.__handlers):
            if component_id == f"handler_{handler_index}":
                return handler
        return None


DynamicWidgetConstructorFn = typing.Callable[[typing.Any], typing.Optional[Declarative.HandlerLike]]


class DynamicWidget(UserInterface.Widget):
    """Widget which only adds content when the produce when produce_widget is called."""
    def __init__(self, ui: UserInterface.UserInterface, event_loop: asyncio.AbstractEventLoop, component_fn: DynamicWidgetConstructorFn) -> None:
        self.__ui = ui
        self.__event_loop = event_loop
        self.__component_fn = component_fn
        self.__widget = ui.create_column_widget()
        super().__init__(Widgets.CompositeWidgetBehavior(self.__widget))
        self.__handler: typing.Optional[Declarative.HandlerLike] = None

    @property
    def dynamic_handler(self) -> typing.Optional[Declarative.HandlerLike]:
        return self.__handler

    def produce_widget(self, item: typing.Any) -> None:
        if self.__widget.child_count == 0:
            self.__handler = self.__component_fn(item)
            if self.__handler:
                self.__widget.add(Declarative.DeclarativeWidget(self.__ui, self.__event_loop, self.__handler))


class DynamicHandler(Declarative.Handler):
    """Dynamic handler which contains a dynamic widget and only makes the widget when produce_widget is called."""
    def __init__(self, item: typing.Any, component_fn: DynamicWidgetConstructorFn) -> None:
        super().__init__()
        self.__item = item
        self.component_fn = component_fn
        self.ui_view = {"type": "dynamic", "name": "dynamic_widget"}
        self.dynamic_widget: typing.Optional[DynamicWidget] = None

    def produce_widget(self) -> None:
        if self.dynamic_widget:
            self.dynamic_widget.produce_widget(self.__item)


class DynamicDeclarativeWidgetConstructor:
    def construct(self, d_type: str, ui: UserInterface.UserInterface, window: typing.Optional[Window.Window],
                  d: Declarative.UIDescription, handler: Declarative.HandlerLike,
                  finishes: typing.List[typing.Callable[[], None]]) -> typing.Optional[UserInterface.Widget]:
        if d_type == "dynamic":
            assert window
            widget = DynamicWidget(ui, window.event_loop, typing.cast(DynamicHandler, handler).component_fn)
            if "name" in d:
                setattr(handler, d["name"], widget)
            return widget
        return None


Registry.register_component(DynamicDeclarativeWidgetConstructor(), {"declarative_constructor"})


_T = typing.TypeVar('_T')

class MappedPropertyListModel(Observable.Observable, typing.Generic[_T]):
    """A list model based on a property of items in a container.

    The list model observes the container for item insertions and removals, reads a property from each item, and
     forms a new list based on those properties. Also observes each item for changes to that property and updates
     this list via both notify_item_content_changed and notify_property_changed('items') when a property changes.
    """

    def __init__(self, container: Observable.Observable, property_name: str, *, parent_items_key: typing.Optional[str] = None, items_key: typing.Optional[str] = None, property_value_fallback: typing.Optional[_T] = None) -> None:
        super().__init__()
        self.__container = container
        self.__property_name = property_name
        self.__property_value_fallback = property_value_fallback
        self.__parent_items_key = parent_items_key or "items"
        self.__items_key = items_key or self.__parent_items_key
        self.__mutex = threading.RLock()
        self.__items: typing.List[typing.Any] = list()  # a list of transformed items
        self.__item_property_values: typing.List[_T] = list()  # a list of transformed items
        self.__item_inserted_event_listener = self.__container.item_inserted_event.listen(ReferenceCounting.weak_partial(MappedPropertyListModel.__parent_item_inserted, self))
        self.__item_removed_event_listener = self.__container.item_removed_event.listen(ReferenceCounting.weak_partial(MappedPropertyListModel.__parent_item_removed, self))
        self.__item_property_changed_event_listeners = list[Event.EventListener]()
        for index, item in enumerate(getattr(self.__container, self.__parent_items_key)):
            self.__parent_item_inserted(self.__parent_items_key, item, index)

    @property
    def items(self) -> typing.Sequence[_T]:
        """ Return the items. """
        with self.__mutex:
            return list(self.__item_property_values)

    def __getattr__(self, property_name: str) -> typing.Any:
        if property_name == self.__items_key:
            return self.items
        raise AttributeError()

    # thread safe.
    def __parent_item_inserted(self, key: str, item: typing.Any, before_index: int) -> None:
        """ Insert the item. Called from the container. """
        if key == self.__parent_items_key:
            with self.__mutex:
                self.__item_property_changed_event_listeners.insert(before_index, item.property_changed_event.listen(ReferenceCounting.weak_partial(MappedPropertyListModel.__item_property_changed, self, item)))
                self.__items.insert(before_index, item)
                self.__item_property_values.insert(before_index, getattr(item, self.__property_name))
                self.notify_insert_item(self.__items_key, item, before_index)
                self.notify_property_changed("items")

    # thread safe.
    def __parent_item_removed(self, key: str, item: typing.Any, index: int) -> None:
        """ Remove the item. Called from the container. """
        if key == self.__parent_items_key:
            with self.__mutex:
                del self.__item_property_changed_event_listeners[index]
                del self.__items[index]
                del self.__item_property_values[index]
                self.notify_remove_item(self.__items_key, item, index)
                self.notify_property_changed("items")

    def __item_property_changed(self, item: typing.Any, property_name: str) -> None:
        """ Called when a property of an item changes. """
        with self.__mutex:
            index = self.__items.index(item)
            self.__item_property_values[index] = getattr(item, self.__property_name)
            self.notify_item_content_changed(self.__items_key, item, index)
            self.notify_property_changed("items")


class MasterDetailHandler(Declarative.Handler):

    def __init__(self, model: Observable.Observable, items_key: str, component_fn: DynamicWidgetConstructorFn,
                 title_property: str, title_fallback: typing.Optional[str], list_props: dict[str, typing.Any], *,
                 add_item_fn: typing.Optional[typing.Callable[[], None]] = None,
                 remove_item_fn: typing.Optional[typing.Callable[[typing.Any], None]] = None) -> None:
        super().__init__()

        self.__component_fn = component_fn

        self.__add_item_fn = add_item_fn
        self.__remove_item_fn = remove_item_fn

        # the items for which the details are displayed.
        self.items_model = model
        self.__items_key = items_key

        # the labels property model converts the titles model to a single property: a list of strings.
        self.labels_property_model = MappedPropertyListModel(self.items_model, title_property, property_value_fallback=title_fallback)

        # set up a shadow list following the model/items_key

        def make_shadow(item: typing.Any) -> typing.Any:
            return DynamicHandler(item, component_fn)

        self.shadow_items = ListModel.MappedListModel(container=self.items_model,
                                                      master_items_key=items_key,
                                                      map_fn=make_shadow)

        # the selected item in the items_model
        self.index_model = Model.PropertyModel(0 if self.__items else None)

        u = Declarative.DeclarativeUI()

        # the list box wants to bind to a property that returns a list of strings. this is the labels property model.
        item_list = u.create_list_box(items_ref="@binding(labels_property_model.items)",
                                      current_index="@binding(index_model.value)",
                                      **list_props)

        item_stack = u.create_stack(items=f"items_model.{items_key}", item_component_id="detail", current_index="@binding(index_model.value)")

        item_control_row_items = list[Declarative.UIDescription]()

        if add_item_fn:
            add_button = u.create_push_button(text=_("+"), on_clicked="add_item", style="minimal", width=24)
            item_control_row_items.append(add_button)

        if remove_item_fn:
            remove_button = u.create_push_button(text=_("-"), on_clicked="remove_item", style="minimal", width=24)
            item_control_row_items.append(remove_button)

        if item_control_row_items:
            item_control_row = u.create_row(u.create_row(*item_control_row_items, spacing=8), u.create_stretch())
            item_list_column = u.create_column(item_list, item_control_row, spacing=8)
        else:
            item_list_column = u.create_column(item_list)

        self.ui_view = u.create_row(item_list_column, u.create_column(item_stack), spacing=12)

        self._detail_components: typing.Dict[typing.Any, typing.Optional[Declarative.HandlerLike]] = dict()

        def index_changed(property: str) -> None:
            if property == "value":
                self.__update_dynamic_widget()

        self.__listener = self.index_model.property_changed_event.listen(index_changed)

    def __update_dynamic_widget(self) -> None:
        index = self.index_model.value or 0
        if 0 <= index < len(self.shadow_items.items):
            typing.cast(DynamicHandler, self.shadow_items.items[index]).produce_widget()

    @property
    def __items(self) -> typing.Sequence[typing.Any]:
        return typing.cast(typing.Sequence[typing.Any], getattr(self.items_model, self.__items_key))

    def add_item(self, widget: UserInterface.Widget) -> None:
        if callable(self.__add_item_fn):
            self.__add_item_fn()
            items = self.__items
            self.index_model.value = len(items) - 1 if items else None

    def remove_item(self, widget: UserInterface.Widget) -> None:
        if callable(self.__remove_item_fn):
            index = self.index_model.value
            items = self.__items
            if index is not None and 0 <= index < len(items):
                self.__remove_item_fn(items[index])
                item_count = len(self.__items)
                if index < item_count:
                    self.index_model.value = None
                    self.index_model.value = index
                elif item_count:
                    self.index_model.value = item_count - 1
                else:
                    self.index_model.value = None

    def init_handler(self) -> None:
        self.__update_dynamic_widget()

    def create_handler(self, component_id: str, container: typing.Optional[ListModel.ListModel[Declarative.HandlerLike]] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        if component_id == "detail":
            assert container
            handler = typing.cast(Declarative.HandlerLike, self.shadow_items.items[container.items.index(item)])
            self._detail_components[item] = handler
            return handler
        return None


class EntityBrowserEntry(Observable.Observable):
    def __init__(self, context: Context, title: str, document_model: Observable.Observable,
                 master_items_key: str, entity_type: typing.Optional[Schema.EntityType], title2: str,
                 title_property: str, title_fallback: typing.Optional[str] = None) -> None:
        super().__init__()
        model = ListModel.FilteredListModel(container=document_model, master_items_key=master_items_key)
        model.sort_key = operator.attrgetter("modified")
        model.sort_reverse = True
        self.model = model
        self.items_key = "items"

        def create_handler(item: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
            used_entity_type = entity_type if entity_type else (item.entity.entity_type if getattr(item, "entity", None) else None)
            return ItemPageHandler(context, item, used_entity_type, title2)

        self.component_fn = create_handler
        self.title_property = title_property
        self.title_fallback = title_fallback
        self.label = title


def make_master_detail(project_item_handler: EntityBrowserEntry) -> Declarative.HandlerLike:
    list_props = {"width": 160, "min_height": 480, "size_policy_vertical": "expanding"}
    return MasterDetailHandler(project_item_handler.model, project_item_handler.items_key, project_item_handler.component_fn, project_item_handler.title_property, project_item_handler.title_fallback, list_props)


class TopLevelItemHandler(Declarative.Handler):
    # a master-detail for each top level item

    def __init__(self, item: typing.Any) -> None:
        super().__init__()
        self.__item = item
        u = Declarative.DeclarativeUI()
        self.ui_view = u.create_component_instance(identifier="content")
        self._master_detail_handler = make_master_detail(self.__item)

    def create_handler(self, component_id: str, container: typing.Optional[typing.Any] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        if component_id == "content":
            return self._master_detail_handler
        return None


def open_project_item(entity_browser_component: Declarative.HandlerLike, base_item: typing.Any, base_type: Schema.EntityType, base_name: str, item: typing.Any) -> None:
    """In the entity browser component, focus the item, contained within base_item having type base_type."""

    entity_browser_component = typing.cast(MasterDetailHandler, entity_browser_component)

    class Visitor(Schema.Visitor):
        def __init__(self, base_value: typing.Any) -> None:
            self.base_value = base_value
            self.breadcrumbs: typing.Sequence[typing.Any] = list()

        def visit(self, accessor: Schema.Accessor) -> None:
            value = accessor.get_value(self.base_value)
            if value == item and isinstance(accessor.field_type, Schema.ComponentType):
                self.breadcrumbs = accessor.breadcrumbs(self.base_value)

    visitor = Visitor(base_item)
    base_type.visit(typing.cast(Schema.Entity, base_item), Schema.BaseAccessor(Schema.reference(base_type), base_name),
                    visitor)

    list_model = typing.cast(ListModel.ListModel[EntityBrowserEntry], entity_browser_component.items_model)
    for index, project_items in enumerate(list_model.items):
        for value in reversed(visitor.breadcrumbs):
            if value in project_items.model.items:
                entity_browser_component.index_model.value = index
                item_index = project_items.model.items.index(value)
                detail_component = typing.cast(DynamicHandler, entity_browser_component._detail_components[
                    list_model._items[entity_browser_component.index_model.value]])
                assert detail_component.dynamic_widget and detail_component.dynamic_widget.dynamic_handler
                content_handler = typing.cast(TopLevelItemHandler, detail_component.dynamic_widget.dynamic_handler)
                detail_handler = typing.cast(MasterDetailHandler, content_handler._master_detail_handler)
                detail_handler.index_model.value = item_index
                return


def make_entity_browser_component(items: typing.Sequence[EntityBrowserEntry]) -> Declarative.HandlerLike:
    """Make an entity browser component with the top level items."""

    list_model = ListModel.ListModel[EntityBrowserEntry](items=items)
    list_props = {"width": 160, "min_height": 480, "size_policy_vertical": "expanding"}
    return MasterDetailHandler(list_model, "items", typing.cast(DynamicWidgetConstructorFn, TopLevelItemHandler), "label", None, list_props)
